import json
from datetime import datetime
import re
import pypinyin
import pymongo
import random
import fitz
import requests
import oss2
import hashlib
from scrapy.utils.python import to_bytes
import base64
import json
import requests


def base64_api(url, uname, pwd, img_bin, typeid):
    """   
    typeid:
    一、图片文字类型(默认 3 数英混合)：
    1 : 纯数字
    1001：纯数字2
    2 : 纯英文
    1002：纯英文2
    3 : 数英混合
    1003：数英混合2
     4 : 闪动GIF
    7 : 无感学习(独家)
    11 : 计算题
    1005:  快速计算题
    16 : 汉字
    32 : 通用文字识别(证件、单据)
    66:  问答题
    49 :recaptcha图片识别
    二、图片旋转角度类型：
    29 :  旋转类型
    三、图片坐标点选类型：
    19 :  1个坐标
    20 :  3个坐标
    21 :  3 ~ 5个坐标
    22 :  5 ~ 8个坐标
    27 :  1 ~ 4个坐标
    48 : 轨迹类型
    四、缺口识别
    18 : 缺口识别（需要2张图 一张目标图一张缺口图）
    33 : 单缺口识别（返回X轴坐标 只需要1张图）
    五、拼图识别
    53：拼图识别
"""
    # with open(img, 'rb') as f:
    #     base64_data = base64.b64encode(f.read())
    #     b64 = base64_data.decode()
    base64_data = base64.b64encode(img_bin)
    b64 = base64_data.decode()
    data = {"username": uname, "password": pwd, "typeid": typeid, "image": b64}
    result = json.loads(requests.post(url, json=data).text)
    if result['success']:
        return result["data"]["result"]
    else:
        print(result["message"])

from scrapy.utils.conf import get_config
from scrapy.exceptions import NotConfigured
def upload_local_pdf(path,aim_file_path): 
    if not get_config().get(section="oss_cfg",option="OSS_USER" ):# 提醒设置参数以及设置默认值
        raise NotConfigured("At scrapy.cfg missing 'OSS_USER'")
    if not get_config().get(section="oss_cfg",option="OSS_PASSWORD" ):
        raise NotConfigured("At scrapy.cfg missing 'OSS_PASSWORD'")
    if not get_config().get(section="oss_cfg",option="FILES_STORE" ):
        raise NotConfigured("At scrapy.cfg missing 'FILES_STORE'")
    OSS_USER = get_config().get(section="oss_cfg",option="OSS_USER" )
    OSS_PASSWORD = get_config().get(section="oss_cfg",option="OSS_PASSWORD" )
    auth = oss2.Auth(OSS_USER, OSS_PASSWORD)
    bucket = oss2.Bucket(auth, 'oss-cn-shenzhen.aliyuncs.com', 'xcc2', enable_crc=False)
    file_store = get_config().get(section="oss_cfg",option="FILES_STORE" )
    media_guid = hashlib.sha1(to_bytes(path)).hexdigest()
    new_url = '{}/{}/{}.{}'.format(file_store, aim_file_path, media_guid, 'pdf')
    try:
        with open(path, 'rb') as f:
            # f.seek(0, os.SEEK_SET)
            # current = f.tell()
            bucket.put_object('{}/{}.{}'.format(aim_file_path, media_guid, 'pdf'), f)
    except:
        new_url = ''
    return new_url


def save_img(url, name):
    resp = requests.get(url)
    path = 'Material\imgs' + name
    with open(path, 'wb') as f:
        f.write(resp.content)
    return path


def img_to_pdf(img_path, img_type):
    # file_name = os.path.basename(img_path).replace(img_type, 'pdf')
    file_name = img_path.replace(img_type, 'pdf')
    doc = fitz.open()
    imgdoc = fitz.open(img_path)
    pdfbytes = imgdoc.convertToPDF()
    imgpdf = fitz.open("pdf", pdfbytes)
    doc.insertPDF(imgpdf)
    doc.save(file_name) 
    doc.close()
    return file_name


def parse_ofps(ofps, ofps_):
    # baike key特征
    overviews_ = ['概述', '产品介绍', 'overview', 'overviews']
    features_ = ['特点', '特性', '产品特点', 'feature', 'features']
    applications_ = ['应用', '用途', '应用场景', 'application', 'applications']
    if ofps:
        overviews = []
        features = []
        applications = []
        fps = [re.sub(r'\r|\t|\n|\s+', ' ', fp.strip()) for fp in ofps if fp.strip()]
        fps = [fp.strip().strip('*').strip('◆').strip('·').strip('●').strip('•').strip('■').strip(':').strip('').strip('★').strip('Ø').strip() for fp in ofps if fp]
        fps = [fp for fp in fps if fp]
        on = False
        fn = False
        an = False
        ind0 = 0
        ind1 = 0
        ind2 = 0
        ind3 = 0
        ind4 = 0
        ind5 = 0
        for i, fp in enumerate(fps, start=1):
            fp_ = re.sub(r'\r|\t|\n|\s+', '', fp.lower().strip())
            if (not overviews) and any([True if f == fp_  else False for f in overviews_]):
                ind0 = i
            if (not on) and any([True if f == fp_ else False for f in features_]):
                on = True
                ind1 = i
            if (not features) and any([True if f == fp_ else False for f in features_]):
                ind2 = i
            if (not fn) and any([True if f == fp_ else False for f in applications_]):
                fn = True
                ind3 = i
            if (not applications) and any([True if f == fp_ else False for f in applications_]):
                ind4 = i
            if (not an) and any([True if f == fp_ else False for f in ofps_]):
                an = True
                ind5 = i
        inds1 = [ind for ind in [ind1, ind3, ind5] if ind > 0]
        ind1 = min(inds1) if inds1 else 0
        inds3 = [ind for ind in [ind3, ind5] if ind > 0]
        ind3 = min(inds3) if inds3 else 0
        overviews = fps[ind0:ind1-1] if ind0 and ind1 else fps[ind0:] if ind0 and (not ind1) else fps
        features = fps[ind2:ind3-1] if ind2 and ind3 else fps[ind2:] if ind2 and (not ind3) else []
        applications = fps[ind4:ind5-1] if ind4 and ind5 else fps[ind4:] if ind4 and (not ind5) else []
        overview = '<br>'.join(overviews) if overviews else ''
        feature = '<br>'.join(features) if features else ''
        application = '<br>'.join(applications) if applications else ''
    else:
        overview = ''
        feature = ''
        application = ''
    return overview, feature, application


def parse_list_json_value(host, tr, rowspan_lists, tr_ind, ut=0):
    # ut: 0-text  1-href  2-text|||href
    values = []
    tds = tr.xpath('./*')
    for td in tds:
        href_x = td.xpath('.//a/@href | .//button/@onclick').extract_first()
        href_r = re.search(r'\"(.*)\"', href_x) if href_x else ''
        href = href_r.group(1) if href_r else href_x.strip() if href_x else ''
        src = td.xpath('.//img/@src').extract_first()
        td_rowspan = td.xpath('./@rowspan').extract_first()
        if href and ('javascript' not in href) and (not src):
            url = href if re.search(r'http', href) else host + href.strip('..')
            text = ' '.join([re.sub(r'\r|\t|\n|\s+', ' ', t) for t in deal_list_space(td.xpath('.//text()').extract()) if t])
            if ut == 2:
                val = (text + '|||' + url)
            elif ut == 1:
                val = url
            else:
                val = text
        elif src:
            val = src if re.search(r'http', src) else host + src.strip('..')
        else:
            val = ' '.join([re.sub(r'\r|\t|\n|\s+', ' ', t) for t in deal_list_space(td.xpath('.//text()').extract()) if t])
            # if td_rowspan:
            #     val = ''.join([t.replace('<br>', ' ') for t in deal_list_space(td.xpath('.//text()').extract())])
            # else:
            #     val = '|'.join([t.replace('<br>', ' ') for t in deal_list_space(td.xpath('.//text()').extract())])
            from Team_Public.configs import REP_VUL
            if val in REP_VUL:
                val = ''
        values.append(val)
        if td_rowspan:
            td_ind = tds.index(td)
            num = int(td_rowspan)
            rowspan_lists.append((tr_ind, td_ind, num, val))
    for rowspan_tuple in rowspan_lists:
        raw_tr_ind = rowspan_tuple[0]
        raw_td_ind = rowspan_tuple[1]
        num = rowspan_tuple[2]
        val = rowspan_tuple[3]
        if raw_tr_ind < tr_ind < num + raw_tr_ind:
            values.insert(raw_td_ind, val)
            # if raw_td_ind == 0:
            #     values.insert(0, val)
            # else:
            #     values.insert(raw_td_ind - 1, val)
    return values


def parse_list_json_key(td_reg, text_reg, tr1, tr2=None, key2=None):
    """" 解析list_json key """
    key1 = []
    key2 = [] if not key2 else key2
    # tr1 = table.xpath('./tr[1]')
    # tr2 = table.xpath('./tr[2]')
    tds1 = tr1.xpath(td_reg)
    tds2 = tr2.xpath(td_reg) if tr2 else []
    for td1 in tds1:
        src = td1.xpath('.//img/@src').extract_first()
        if src:
            td_text = src
        else:
            try:
                td_text = td1.xpath(text_reg).extract()
            except:
                td_text = []
            if not td_text:
                try:
                    text_reg = text_reg[:text_reg.rfind('/')]
                    td_text = td1.xpath(text_reg).extract()
                except:
                    td_text = []
            td_text = ' '.join(deal_list_space(td_text))
            # td_text = re.sub(r'\r|\t|\n|((?!<small[^>]*>).)+|((?!<sup[^>]*>).)+|((?!<sub[^>]*>).)+', '', td_text) if td_text else ''
            td_text = re.sub(r'\r|\t|\n|\s+|<(?!(small|/small|sub|/sub|sup|/sup))[^>]*>', ' ', td_text) if td_text else ''
        td_text = td_text.strip()
        col_span = td1.xpath('./@colspan').extract_first()
        row_span = td1.xpath('./@rowspan').extract_first()
        if col_span and int(col_span) > 1:
            key1.append([td_text] * int(col_span))
        elif not row_span or int(row_span) < 2:
            key1.append([td_text])
        else:
            key1.append(td_text)
    for td2 in tds2:
        try:
            td_text = td2.xpath(text_reg).extract()
        except:
            td_text = []
        if not td_text:
            try:
                text_reg = text_reg[:text_reg.rfind('/')]
                td_text = td2.xpath(text_reg).extract()
            except:
                td_text = []
        td_text = ' '.join(deal_list_space(td_text))
        # td_text = td2.xpath(text_reg).xpath('string(.)').extract_first()
        td_text = re.sub(r'\r|\t|\n|\s+|<(?!(small|/small|sub|/sub|sup|/sup))[^>]*>', ' ', td_text) if td_text else ''
        td_text = td_text.strip()
        key2.append(td_text)
    for i, k1 in enumerate(key1):
        ind = i
        if isinstance(k1, list):
            k2 = key2[:len(k1)]
            ks = lists_add((k1, k2))
            key1[ind] = ks
            key2 = key2[len(k2):]
            # for i in range(len(k2)+1):
            #     key2.pop(i)
    keys = []
    for key in key1:
        if isinstance(key, list):
            for k in key:
                keys.append(k)
        else:
            keys.append(key)
    return keys


def parse_list_json_keys(tbody, keys_rowspan, tds_reg, key_text_reg, k):
    if keys_rowspan in [1]:
        k_tds = tbody.xpath('./tr[{}]'.format(keys_rowspan + k)).xpath(tds_reg)
        keys = []
        for k_td in k_tds:
            # text = k_td.xpath(key_text_reg).xpath('string(.)').extract_first()
            try:
                text = k_td.xpath(key_text_reg).extract()
            except:
                text = []
                print(k_td, key_text_reg)
            if not text:
                try:
                    key_text_reg = key_text_reg[:key_text_reg.rfind('/')]
                    text = k_td.xpath(key_text_reg).extract()
                except:
                    text = []
            # td_text = ' '.join(deal_list_space(text))
            td_text = [t.strip() for t in text]
            td_text = ' '.join([tt for tt in td_text if tt])
            # key = re.sub(r'\r|\t|\n|<br[^>]*>|<p[^>]*>|</p>', ' ', td_text) if td_text else ''
            # key = re.sub(r'<(?!(small|/small|sub|/sub|sup|/sup))[^>]*>', '', key) if key else ''
            key = re.sub(r'\r|\t|\n|\s+|<(?!(small|/small|sub|/sub|sup|/sup))[^>]*>', ' ', td_text) if td_text else ''
            keys.append(key.strip())
        # tr1 = tbody.xpath('./tr[{}]'.format(keys_rowspan + k))
        # keys = parse_list_json_key(tds_reg, key_text_reg, tr1)
    elif keys_rowspan in [2]:
        tr1 = tbody.xpath('./tr[{}]'.format(keys_rowspan - 1 + k))
        tr2 = tbody.xpath('./tr[{}]'.format(keys_rowspan + k))
        keys = parse_list_json_key(tds_reg, key_text_reg, tr1, tr2)
    else:
        tr1 = tbody.xpath('./tr[{}]'.format(keys_rowspan + k))
        tr2 = tbody.xpath('./tr[{}]'.format(keys_rowspan - 1 + k))
        keys = parse_list_json_key(tds_reg, key_text_reg, tr1, tr2)
        for kr in range(keys_rowspan - 2, 0, -1):
            tr1 = tbody.xpath('./tr[{}]'.format(kr + k))
            keys = parse_list_json_key(tds_reg, key_text_reg, tr1, None, keys)
    return keys


def category_item(response, config, item, category, level=None, p_category_name=None, p_category_id=None):
    if level:
        item['level'] = level
    if p_category_name: item['p_category_name'] = p_category_name
    if p_category_id: item['p_category_id'] = p_category_id
    # category_name = p_category_name + '_' + category if p_category_name else category
    category_name = category
    category_id = pinyin(category_name, config)
    item['category_name'] = category_name
    item['category_id'] = category_id
    item['url'] = response.url
    item = item_constant(item, config)
    return item


def item_constant(item, config):
    # item["table"] = config.TABLE
    # item["sources"] = config.SOURCES
    # item["creator"] = config.CREATOR
    # item["create_time"] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    # item['brand_name'] = config.BRAND_NAME
    # item['brand_id'] = config.BRAND_ID
    item["sources"] = config.get('sources')
    item["creator"] = config.get('creator')
    item["create_time"] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    item['brand_name'] = config.get('brand_name')
    item['brand_id'] = config.get('brand_id')
    return item


# def up_config(config, brand_name, brand_id, creator, keys, vals, categorys):
#     config.SOURCES = 'manufacturers'
#     config.BRAND_NAME = brand_name
#     config.BRAND_ID = brand_id
#     config.CREATOR = creator
#     #  过滤list_json字段
#     config.DEL_KEY = keys
#     config.DEL_VUL = vals
#     # category过滤
#     config.categorys_names = categorys
#     return config



def mongodb_find_max_date(table_name, filte_dict, sort_str):
    mongocli = pymongo.MongoClient('10.8.108.201', 27018)
    db = mongocli.spider
    db.authenticate("spider", "Dashuju.spider")
    dbcol = db[table_name]
    max_date = ''
    for x in dbcol.find(filte_dict, {sort_str: 1}).sort([(sort_str, -1)]).skip(0).limit(1):
        res = x.get(sort_str)
        if res:
            max_date = res
            break
    return max_date


def lists_add(lists):
    """ 多list相加 """
    results = []
    for tr_text in zip(*lists):
        k = ''
        for tr in tr_text:
            if str(tr) != k.strip():
                k += ' ' + str(tr)
        results.append(k.strip())
    return results


def filter_dict(dicts, config):
    """ 删除dict指定key与value """
    del_key = [k for k in config.get('del_key')]
    del_val = [v for v in config.get('del_val')]
    vals = list(dicts.values())
    keys = list(dicts.keys())
    # key_lower = [k.lower() for k in list(dicts.keys())]
    for i in range(len(vals)):
        if vals[i] in del_val:
            if keys[i] not in del_key:
                del_key.append(keys[i])
    for kk in del_key:
        for key in keys:
            # if kk.lower() in key.lower():
            if kk.lower() == key.lower():
                try:
                    dicts.pop(key)
                    break
                except:
                    continue
    return dicts


def pinyin(word, config):
    """ 汉字转拼音 """
    if (not config.get('brand_name')) or not word: return ''
    result = ''
    for i in pypinyin.pinyin(config.get('brand_name') + '_' + word, style=pypinyin.NORMAL):
        result += ''.join(i)
    return result


def deal_space(lists):
    """ 列表空格删除处理 """
    list_new = []
    for li in lists:
        li_new = re.sub(r'\s|\t|\n', '', li)
        if li_new:
            list_new.append(li_new)
    return list_new


def deal_lines(response, xpath_reg, i):
    res_list = []
    while True:
        results = response.xpath(xpath_reg.format(i)).extract()
        results = deal_space(results)
        if not results:
            break
        res_list += results
        i += 1
    data_list = []
    for res in res_list:
        data_list.append('<p>{}</p>'.format(res))
    return ''.join(data_list)


def deal_list_space(lists):
    """ 列表元素前后空格删除 """
    list_new = []
    for li in lists:
        li = li.strip()
        if li:
            list_new.append(li)
    return list_new


def deal_list_tag(lists, reg=None):
    """ 去除列表元素标签 """
    list_new = []
    for li in lists:
        new_val = re.search(reg, li)
        if new_val:
            list_new.append(new_val.group(1).strip())
    return list_new


def json_dump(item, file):
    datas = json.dumps(dict(item), indent=2, separators=(',', ':'), ensure_ascii=False)
    with open(file, 'a+', encoding='utf-8') as f:
        f.write(datas + '\n')



def obsolete_setter(setter, attrname):
    def newsetter(self, value):
        c = self.__class__.__name__
        msg = f"{c}.{attrname} is not modifiable, use {c}.replace() instead"
        raise AttributeError(msg)
    return newsetter



# -*- coding: utf-8 -*-
"""
更新日期:2022/4/28
作者: D.Z.zhong
用途: 图片管道文件实现异步下载上传至阿里云oss
运行环境：win10 64/linux + python3.9 + scrapy2.4.1 + oss2 + magic(pip install python-magic-bin)
选择目录位置:{python}/Lib/site-packages/scrapy/xcc_pipelines
"""

import logging
from io import BytesIO
from urllib.parse import urlparse

from .ossfilepipelines import OssFilesPipeline
from scrapy.utils.request import referer_str
from scrapy.pipelines.files import FileException

from scrapy.utils.misc import arg_to_iter
from twisted.internet.defer import DeferredList

# logging.basicConfig(level=logging.INFO)
logging.getLogger('pdfplumber').setLevel(logging.WARNING)
logging.getLogger('PyPDF2').setLevel(logging.WARNING)
logging.getLogger('pdfminer').setLevel(logging.WARNING)
logging.getLogger('urllib3').setLevel(logging.WARNING)
logging.getLogger('oss2').setLevel(logging.WARNING)
logging.getLogger('PIL').setLevel(logging.WARNING)
logger = logging.getLogger(__name__)




class OssImagesPipeline(OssFilesPipeline): #图片类
    DEFAULT_FILES_OSS_URLS_FIELD = 'raw_img_url'
    DEFAULT_FILES_OSS_RESULT_FIELD = 'img_url'
    MIN_WIDTH = 50
    MIN_HEIGHT = 50
    def process_item(self, item, spider):
        info = self.spiderinfo
        requests = arg_to_iter(self.get_media_requests(item, info))
        dlist = [self._process_request(r, info, item) for r in requests]
        dfd = DeferredList(dlist, consumeErrors=1)
        return dfd.addCallback(self.item_completed, item, info)

    def media_downloaded(self, response, request, info, *, item=None):
        '''多处理了一步图片大小判断'''
        from PIL import Image
        orig_image = Image.open(BytesIO(response.body))

        width, height = orig_image.size
        if width < self.MIN_WIDTH or height < self.MIN_HEIGHT:
            raise ImageException("Image too small "
                                 f"({width}x{height} < "
                                 f"{self.MIN_WIDTH}x{self.MIN_HEIGHT})")

        referer = referer_str(request)
        if response.status != 200:
            logger.warning(
                'File (code: %(status)s): Error downloading file from '
                '%(request)s referred in <%(referer)s>',
                {'status': response.status,
                 'request': request, 'referer': referer},
                extra={'spider': info.spider}
            )
            raise FileException('download-error')

        if not response.body:
            logger.warning(
                'File (empty-content): Empty file from %(request)s referred '
                'in <%(referer)s>: no-content',
                {'request': request, 'referer': referer},
                extra={'spider': info.spider}
            )
            raise FileException('empty-content')

        status = 'cached' if 'cached' in response.flags else 'downloaded'
        logger.debug(
            'File (%(status)s): Downloaded file from %(request)s referred in '
            '<%(referer)s>',
            {'status': status, 'request': request, 'referer': referer},
            extra={'spider': info.spider}
        )
        self.inc_stats(info.spider, status)

        try:
            path = self.file_path(request, response=response, info=info, item=item)
            checksum = self.file_downloaded(response, request, info, item=path)
        except FileException as exc:
            logger.warning(
                'File (error): Error processing file from %(request)s '
                'referred in <%(referer)s>: %(errormsg)s',
                {'request': request, 'referer': referer, 'errormsg': str(exc)},
                extra={'spider': info.spider}, exc_info=True
            )
            raise
        except Exception as exc:
            logger.error(
                'File (unknown-error): Error processing file from %(request)s '
                'referred in <%(referer)s>',
                {'request': request, 'referer': referer},
                exc_info=True, extra={'spider': info.spider}
            )
            raise FileException(str(exc))

        return {'url': request.url, 'path': path, 'checksum': checksum, 'status': status}
        # return super().media_downloaded(response, request, info, item=item)

class OssImagesPipelineBak(OssImagesPipeline): #图片类备用 如电路图
    pass


class ImageException(FileException):
    """General image error exception"""


import pickle
import sys

from queuelib.tests import test_queue as t
from scrapy.squeues import (
    _MarshalFifoSerializationDiskQueue,
    _MarshalLifoSerializationDiskQueue,
    _PickleFifoSerializationDiskQueue,
    _PickleLifoSerializationDiskQueue,
)
from scrapy.item import Item, Field
from scrapy.http import Request
from scrapy.loader import ItemLoader
from scrapy.selector import Selector


class TestItem(Item):
    name = Field()


def _test_procesor(x):
    return x + x


class TestLoader(ItemLoader):
    default_item_class = TestItem
    name_out = staticmethod(_test_procesor)


def nonserializable_object_test(self):
    q = self.queue()
    self.assertRaises(ValueError, q.push, lambda x: x)
    # Selectors should fail (lxml.html.HtmlElement objects can't be pickled)
    sel = Selector(text='<html><body><p>some text</p></body></html>')
    self.assertRaises(ValueError, q.push, sel)


class FifoDiskQueueTestMixin:

    def test_serialize(self):
        q = self.queue()
        q.push('a')
        q.push(123)
        q.push({'a': 'dict'})
        self.assertEqual(q.pop(), 'a')
        self.assertEqual(q.pop(), 123)
        self.assertEqual(q.pop(), {'a': 'dict'})

    test_nonserializable_object = nonserializable_object_test


class MarshalFifoDiskQueueTest(t.FifoDiskQueueTest, FifoDiskQueueTestMixin):
    chunksize = 100000

    def queue(self):
        return _MarshalFifoSerializationDiskQueue(self.qpath, chunksize=self.chunksize)


class ChunkSize1MarshalFifoDiskQueueTest(MarshalFifoDiskQueueTest):
    chunksize = 1


class ChunkSize2MarshalFifoDiskQueueTest(MarshalFifoDiskQueueTest):
    chunksize = 2


class ChunkSize3MarshalFifoDiskQueueTest(MarshalFifoDiskQueueTest):
    chunksize = 3


class ChunkSize4MarshalFifoDiskQueueTest(MarshalFifoDiskQueueTest):
    chunksize = 4


class PickleFifoDiskQueueTest(t.FifoDiskQueueTest, FifoDiskQueueTestMixin):

    chunksize = 100000

    def queue(self):
        return _PickleFifoSerializationDiskQueue(self.qpath, chunksize=self.chunksize)

    def test_serialize_item(self):
        q = self.queue()
        i = TestItem(name='foo')
        q.push(i)
        i2 = q.pop()
        assert isinstance(i2, TestItem)
        self.assertEqual(i, i2)

    def test_serialize_loader(self):
        q = self.queue()
        loader = TestLoader()
        q.push(loader)
        loader2 = q.pop()
        assert isinstance(loader2, TestLoader)
        assert loader2.default_item_class is TestItem
        self.assertEqual(loader2.name_out('x'), 'xx')

    def test_serialize_request_recursive(self):
        q = self.queue()
        r = Request('http://www.example.com')
        r.meta['request'] = r
        q.push(r)
        r2 = q.pop()
        assert isinstance(r2, Request)
        self.assertEqual(r.url, r2.url)
        assert r2.meta['request'] is r2

    def test_non_pickable_object(self):
        q = self.queue()
        try:
            q.push(lambda x: x)
        except ValueError as exc:
            if hasattr(sys, "pypy_version_info"):
                self.assertIsInstance(exc.__context__, pickle.PicklingError)
            else:
                self.assertIsInstance(exc.__context__, AttributeError)
        sel = Selector(text='<html><body><p>some text</p></body></html>')
        try:
            q.push(sel)
        except ValueError as exc:
            self.assertIsInstance(exc.__context__, TypeError)


class ChunkSize1PickleFifoDiskQueueTest(PickleFifoDiskQueueTest):
    chunksize = 1


class ChunkSize2PickleFifoDiskQueueTest(PickleFifoDiskQueueTest):
    chunksize = 2


class ChunkSize3PickleFifoDiskQueueTest(PickleFifoDiskQueueTest):
    chunksize = 3


class ChunkSize4PickleFifoDiskQueueTest(PickleFifoDiskQueueTest):
    chunksize = 4


class LifoDiskQueueTestMixin:

    def test_serialize(self):
        q = self.queue()
        q.push('a')
        q.push(123)
        q.push({'a': 'dict'})
        self.assertEqual(q.pop(), {'a': 'dict'})
        self.assertEqual(q.pop(), 123)
        self.assertEqual(q.pop(), 'a')

    test_nonserializable_object = nonserializable_object_test


class MarshalLifoDiskQueueTest(t.LifoDiskQueueTest, LifoDiskQueueTestMixin):

    def queue(self):
        return _MarshalLifoSerializationDiskQueue(self.qpath)


class PickleLifoDiskQueueTest(t.LifoDiskQueueTest, LifoDiskQueueTestMixin):

    def queue(self):
        return _PickleLifoSerializationDiskQueue(self.qpath)

    def test_serialize_item(self):
        q = self.queue()
        i = TestItem(name='foo')
        q.push(i)
        i2 = q.pop()
        assert isinstance(i2, TestItem)
        self.assertEqual(i, i2)

    def test_serialize_loader(self):
        q = self.queue()
        loader = TestLoader()
        q.push(loader)
        loader2 = q.pop()
        assert isinstance(loader2, TestLoader)
        assert loader2.default_item_class is TestItem
        self.assertEqual(loader2.name_out('x'), 'xx')

    def test_serialize_request_recursive(self):
        q = self.queue()
        r = Request('http://www.example.com')
        r.meta['request'] = r
        q.push(r)
        r2 = q.pop()
        assert isinstance(r2, Request)
        self.assertEqual(r.url, r2.url)
        assert r2.meta['request'] is r2



"""
date:2021/10/26
auth:t.y.jie
"""
import json
import pymongo
from pymongo import MongoClient,ReadPreference
from scrapy.utils.conf import get_config
import os
import logging
import time
from datetime import datetime

from twisted.internet.threads import deferToThread

class MongodbPipeline(object):
    def __init__(self,MONGO_HOST,MONGO_PORT,MONGO_PSW,MONGO_USER,MONGO_DB,AUTH_SOURCE,SECTION_SELECT):
        # 链接数据库
        if MONGO_PORT:
            # 兼容之前的settings配置
            mongo_url = 'mongodb://{0}:{1}@{2}:{3}/?authSource={4}&authMechanism=SCRAM-SHA-1'.format(MONGO_USER, MONGO_PSW,
                                                                                                     MONGO_HOST, MONGO_PORT,
                                                                                                     AUTH_SOURCE)
            self.client = MongoClient(mongo_url)
            self.db = self.client[MONGO_DB]  # 获得数据库的句柄
        else:
            mongo_url = 'mongodb://{0}:{1}@{2}/?authSource={3}&replicaSet=rs01'.format(MONGO_USER, MONGO_PSW,
                                                                                                     MONGO_HOST,
                                                                                                     AUTH_SOURCE)
            self.client = MongoClient(mongo_url)
            # 读写分离
            self.db = self.client.get_database(MONGO_DB, read_preference=ReadPreference.SECONDARY_PREFERRED)
        print("mongo_url:",mongo_url)
        self.section_select = SECTION_SELECT
    @classmethod
    def from_crawler(cls, crawler):
        # 判断运行环境<根据环境变量中是否配置IF_PROD=True,或 测试正式环境settings["IF_PROD"] = True>
        section_select ="mongo_cfg_prod" if os.environ.get('IF_PROD') == "True" or crawler.settings.get('IF_PROD')\
             == True or get_config().get("settings","IF_PROD",fallback="False")=="True" else "mongo_cfg_dev"
        
        return cls(MONGO_HOST=get_config().get(section=section_select,option='MONGO_HOST',fallback='') or crawler.settings.get('MONGO_HOST'),
                   MONGO_PORT=get_config().getint(section=section_select,option='MONGO_PORT',fallback='') or crawler.settings.get('MONGO_PORT'),
                   MONGO_PSW=get_config().get(section=section_select,option='MONGO_PSW',fallback='') or crawler.settings.get('MONGO_PSW'),
                   MONGO_USER=get_config().get(section=section_select,option='MONGO_USER',fallback='') or crawler.settings.get('MONGO_USER'),
                   MONGO_DB=get_config().get(section=section_select,option='MONGO_DB',fallback='') or crawler.settings.get('MONGO_DB'),
                   AUTH_SOURCE=get_config().get(section=section_select,option='AUTH_SOURCE',fallback='') or crawler.settings.get('AUTH_SOURCE'),
                   SECTION_SELECT = section_select)

    def _process_item(self, item, spider):
        date_time = datetime.now().strftime("_%Y_%m") if self.section_select=='mongo_cfg_prod' else ""
        postItem = dict(item)
        if postItem.get("brand_id", None):  # brand_id :: int
            postItem["brand_id"] = int(postItem["brand_id"])
        if postItem.get("level", None):  # level :: int
            postItem["level"] = int(postItem["level"])
        if postItem.get("list_json", None):  # list_json :: json
            if isinstance(postItem["list_json"], dict):
                postItem["list_json"] = json.dumps(dict(postItem["list_json"]))
        if postItem.get("min_work_tp",None):  # min_work_tp :: int
            postItem["min_work_tp"] = int(postItem["min_work_tp"])
        if postItem.get("max_work_tp",None):  # max_work_tp :: int
            postItem["max_work_tp"] = int(postItem["max_work_tp"])
        if postItem.get("moq", None):  # moq :: int
            postItem["moq"] = int(postItem["moq"].replace(",",''))
        if postItem.get("mpq", None):  # mpq :: int
            postItem["mpq"] = int(postItem["mpq"].replace(",",''))
        if postItem.get("rough_weight", None):  # rough_weight :: float #注意单位是g
            postItem["rough_weight"] = int(postItem["rough_weight"])
        if postItem.get("title", None):  # title :: 去两端空格
            postItem["title"] = postItem["title"].strip()
        if postItem.get("sources",None):  # sources :: 去两端空格
            postItem["sources"] = postItem["sources"].strip()
        if postItem.get("brand_name", None):  # brand_name :: 去两端空格
            postItem["brand_name"] = postItem["brand_name"].strip()
        if postItem.get("category_id", None):  # category_id :: 去两端空格
            postItem["category_id"] = postItem["category_id"].strip()
        if postItem.get("category_name", None):  # category_name :: 去两端空格
            postItem["category_name"] = postItem["category_name"].strip()
        if postItem.get("p_category_id", None):  # p_category_id :: 去两端空格
            postItem["p_category_id"] = postItem["p_category_id"].strip()
        if postItem.get("p_category_name", None):  # p_category_name :: 去两端空格
            postItem["p_category_name"] = postItem["p_category_name"].strip()
        if postItem.get("brand_id", None):  # brand_id :: int
            postItem["brand_id"] = int(postItem["brand_id"])
        if postItem.get("level", None):  # level :: int
            postItem["level"] = int(postItem["level"])
        if postItem.get("list_json", None):  # list_json :: json
            if isinstance(postItem["list_json"], dict):
                postItem["list_json"] = json.dumps(dict(postItem["list_json"]))
        if postItem.get("min_work_tp",None):  # min_work_tp :: int
            postItem["min_work_tp"] = int(postItem["min_work_tp"])
        if postItem.get("max_work_tp",None):  # max_work_tp :: int
            postItem["max_work_tp"] = int(postItem["max_work_tp"])
        if postItem.get("moq", None):  # moq :: int
            postItem["moq"] = int(postItem["moq"])
        if postItem.get("mpq", None):  # mpq :: int
            postItem["mpq"] = int(postItem["mpq"])
        if postItem.get("rough_weight", None):  # rough_weight :: float #注意:单位g
            postItem["rough_weight"] = float(postItem["rough_weight"])
        if postItem.get('create_time', None):  # create_time标准时间格式2022-01-01 00:00:00
            postItem['create_time'] = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time()))
        if postItem.get('spider_time', None):  # spider_time标准时间格式2022-01-01 00:00:00
            postItem['spider_time'] = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time()))
        msg = postItem.get("title", "") or postItem.get("category_name","") or postItem.get("url","") or postItem.get("sources","") # or postItem.get("")
        try:
             # 把item转化成字典形式
            coll = self.db[item.table+date_time]
            coll.insert(postItem)  # 向数据库插入一条记录
            logging.debug(f'Crawl {msg} done.' )
        except pymongo.errors.DuplicateKeyError:
            logging.info(f'去重 {msg} Skip .') 
        return item

    def process_item(self, item, spider):
        return deferToThread(self._process_item, item, spider)

    def close_spider(self,spider):
        self.client.close()



import hashlib
import io
import random
from shutil import rmtree
from tempfile import mkdtemp
from unittest import skipIf

import attr
from itemadapter import ItemAdapter
from twisted.trial import unittest

from scrapy.http import Request, Response
from scrapy.item import Field, Item
from scrapy.pipelines.images import ImagesPipeline
from scrapy.settings import Settings
from scrapy.utils.python import to_bytes


try:
    from dataclasses import make_dataclass, field as dataclass_field
except ImportError:
    make_dataclass = None
    dataclass_field = None


try:
    from PIL import Image
except ImportError:
    skip_pillow = 'Missing Python Imaging Library, install https://pypi.python.org/pypi/Pillow'
else:
    encoders = {'jpeg_encoder', 'jpeg_decoder'}
    if not encoders.issubset(set(Image.core.__dict__)):
        skip_pillow = 'Missing JPEG encoders'
    else:
        skip_pillow = None


def _mocked_download_func(request, info):
    response = request.meta.get('response')
    return response() if callable(response) else response


class ImagesPipelineTestCase(unittest.TestCase):

    skip = skip_pillow

    def setUp(self):
        self.tempdir = mkdtemp()
        self.pipeline = ImagesPipeline(self.tempdir, download_func=_mocked_download_func)

    def tearDown(self):
        rmtree(self.tempdir)

    def test_file_path(self):
        file_path = self.pipeline.file_path
        self.assertEqual(
            file_path(Request("https://dev.mydeco.com/mydeco.gif")),
            'full/3fd165099d8e71b8a48b2683946e64dbfad8b52d.jpg')
        self.assertEqual(
            file_path(Request("http://www.maddiebrown.co.uk///catalogue-items//image_54642_12175_95307.jpg")),
            'full/0ffcd85d563bca45e2f90becd0ca737bc58a00b2.jpg')
        self.assertEqual(
            file_path(Request("https://dev.mydeco.com/two/dirs/with%20spaces%2Bsigns.gif")),
            'full/b250e3a74fff2e4703e310048a5b13eba79379d2.jpg')
        self.assertEqual(
            file_path(Request("http://www.dfsonline.co.uk/get_prod_image.php?img=status_0907_mdm.jpg")),
            'full/4507be485f38b0da8a0be9eb2e1dfab8a19223f2.jpg')
        self.assertEqual(
            file_path(Request("http://www.dorma.co.uk/images/product_details/2532/")),
            'full/97ee6f8a46cbbb418ea91502fd24176865cf39b2.jpg')
        self.assertEqual(
            file_path(Request("http://www.dorma.co.uk/images/product_details/2532")),
            'full/244e0dd7d96a3b7b01f54eded250c9e272577aa1.jpg')
        self.assertEqual(
            file_path(Request("http://www.dorma.co.uk/images/product_details/2532"),
                      response=Response("http://www.dorma.co.uk/images/product_details/2532"),
                      info=object()),
            'full/244e0dd7d96a3b7b01f54eded250c9e272577aa1.jpg')

    def test_thumbnail_name(self):
        thumb_path = self.pipeline.thumb_path
        name = '50'
        self.assertEqual(thumb_path(Request("file:///tmp/foo.jpg"), name),
                         'thumbs/50/38a86208c36e59d4404db9e37ce04be863ef0335.jpg')
        self.assertEqual(thumb_path(Request("file://foo.png"), name),
                         'thumbs/50/e55b765eba0ec7348e50a1df496040449071b96a.jpg')
        self.assertEqual(thumb_path(Request("file:///tmp/foo"), name),
                         'thumbs/50/0329ad83ebb8e93ea7c7906d46e9ed55f7349a50.jpg')
        self.assertEqual(thumb_path(Request("file:///tmp/some.name/foo"), name),
                         'thumbs/50/850233df65a5b83361798f532f1fc549cd13cbe9.jpg')
        self.assertEqual(thumb_path(Request("file:///tmp/some.name/foo"), name,
                                    response=Response("file:///tmp/some.name/foo"),
                                    info=object()),
                         'thumbs/50/850233df65a5b83361798f532f1fc549cd13cbe9.jpg')

    def test_convert_image(self):
        SIZE = (100, 100)
        # straigh forward case: RGB and JPEG
        COLOUR = (0, 127, 255)
        im = _create_image('JPEG', 'RGB', SIZE, COLOUR)
        converted, _ = self.pipeline.convert_image(im)
        self.assertEqual(converted.mode, 'RGB')
        self.assertEqual(converted.getcolors(), [(10000, COLOUR)])

        # check that thumbnail keep image ratio
        thumbnail, _ = self.pipeline.convert_image(converted, size=(10, 25))
        self.assertEqual(thumbnail.mode, 'RGB')
        self.assertEqual(thumbnail.size, (10, 10))

        # transparency case: RGBA and PNG
        COLOUR = (0, 127, 255, 50)
        im = _create_image('PNG', 'RGBA', SIZE, COLOUR)
        converted, _ = self.pipeline.convert_image(im)
        self.assertEqual(converted.mode, 'RGB')
        self.assertEqual(converted.getcolors(), [(10000, (205, 230, 255))])

        # transparency case with palette: P and PNG
        COLOUR = (0, 127, 255, 50)
        im = _create_image('PNG', 'RGBA', SIZE, COLOUR)
        im = im.convert('P')
        converted, _ = self.pipeline.convert_image(im)
        self.assertEqual(converted.mode, 'RGB')
        self.assertEqual(converted.getcolors(), [(10000, (205, 230, 255))])


class DeprecatedImagesPipeline(ImagesPipeline):
    def file_key(self, url):
        return self.image_key(url)

    def image_key(self, url):
        image_guid = hashlib.sha1(to_bytes(url)).hexdigest()
        return f'empty/{image_guid}.jpg'

    def thumb_key(self, url, thumb_id):
        thumb_guid = hashlib.sha1(to_bytes(url)).hexdigest()
        return f'thumbsup/{thumb_id}/{thumb_guid}.jpg'


class ImagesPipelineTestCaseFieldsMixin:

    skip = skip_pillow

    def test_item_fields_default(self):
        url = 'http://www.example.com/images/1.jpg'
        item = self.item_class(name='item1', image_urls=[url])
        pipeline = ImagesPipeline.from_settings(Settings({'IMAGES_STORE': 's3://example/images/'}))
        requests = list(pipeline.get_media_requests(item, None))
        self.assertEqual(requests[0].url, url)
        results = [(True, {'url': url})]
        item = pipeline.item_completed(results, item, None)
        images = ItemAdapter(item).get("images")
        self.assertEqual(images, [results[0][1]])
        self.assertIsInstance(item, self.item_class)

    def test_item_fields_override_settings(self):
        url = 'http://www.example.com/images/1.jpg'
        item = self.item_class(name='item1', custom_image_urls=[url])
        pipeline = ImagesPipeline.from_settings(Settings({
            'IMAGES_STORE': 's3://example/images/',
            'IMAGES_URLS_FIELD': 'custom_image_urls',
            'IMAGES_RESULT_FIELD': 'custom_images'
        }))
        requests = list(pipeline.get_media_requests(item, None))
        self.assertEqual(requests[0].url, url)
        results = [(True, {'url': url})]
        item = pipeline.item_completed(results, item, None)
        custom_images = ItemAdapter(item).get("custom_images")
        self.assertEqual(custom_images, [results[0][1]])
        self.assertIsInstance(item, self.item_class)


class ImagesPipelineTestCaseFieldsDict(ImagesPipelineTestCaseFieldsMixin, unittest.TestCase):
    item_class = dict


class ImagesPipelineTestItem(Item):
    name = Field()
    # default fields
    image_urls = Field()
    images = Field()
    # overridden fields
    custom_image_urls = Field()
    custom_images = Field()


class ImagesPipelineTestCaseFieldsItem(ImagesPipelineTestCaseFieldsMixin, unittest.TestCase):
    item_class = ImagesPipelineTestItem


@skipIf(not make_dataclass, "dataclasses module is not available")
class ImagesPipelineTestCaseFieldsDataClass(ImagesPipelineTestCaseFieldsMixin, unittest.TestCase):
    item_class = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if make_dataclass:
            self.item_class = make_dataclass(
                "FilesPipelineTestDataClass",
                [
                    ("name", str),
                    # default fields
                    ("image_urls", list, dataclass_field(default_factory=list)),
                    ("images", list, dataclass_field(default_factory=list)),
                    # overridden fields
                    ("custom_image_urls", list, dataclass_field(default_factory=list)),
                    ("custom_images", list, dataclass_field(default_factory=list)),
                ],
            )


@attr.s
class ImagesPipelineTestAttrsItem:
    name = attr.ib(default="")
    # default fields
    image_urls = attr.ib(default=lambda: [])
    images = attr.ib(default=lambda: [])
    # overridden fields
    custom_image_urls = attr.ib(default=lambda: [])
    custom_images = attr.ib(default=lambda: [])


class ImagesPipelineTestCaseFieldsAttrsItem(ImagesPipelineTestCaseFieldsMixin, unittest.TestCase):
    item_class = ImagesPipelineTestAttrsItem


class ImagesPipelineTestCaseCustomSettings(unittest.TestCase):

    skip = skip_pillow

    img_cls_attribute_names = [
        # Pipeline attribute names with corresponding setting names.
        ("EXPIRES", "IMAGES_EXPIRES"),
        ("MIN_WIDTH", "IMAGES_MIN_WIDTH"),
        ("MIN_HEIGHT", "IMAGES_MIN_HEIGHT"),
        ("IMAGES_URLS_FIELD", "IMAGES_URLS_FIELD"),
        ("IMAGES_RESULT_FIELD", "IMAGES_RESULT_FIELD"),
        ("THUMBS", "IMAGES_THUMBS")
    ]

    # This should match what is defined in ImagesPipeline.
    default_pipeline_settings = dict(
        MIN_WIDTH=0,
        MIN_HEIGHT=0,
        EXPIRES=90,
        THUMBS={},
        IMAGES_URLS_FIELD='image_urls',
        IMAGES_RESULT_FIELD='images'
    )

    def setUp(self):
        self.tempdir = mkdtemp()

    def tearDown(self):
        rmtree(self.tempdir)

    def _generate_fake_settings(self, prefix=None):
        """
        :param prefix: string for setting keys
        :return: dictionary of image pipeline settings
        """

        def random_string():
            return "".join([chr(random.randint(97, 123)) for _ in range(10)])

        settings = {
            "IMAGES_EXPIRES": random.randint(100, 1000),
            "IMAGES_STORE": self.tempdir,
            "IMAGES_RESULT_FIELD": random_string(),
            "IMAGES_URLS_FIELD": random_string(),
            "IMAGES_MIN_WIDTH": random.randint(1, 1000),
            "IMAGES_MIN_HEIGHT": random.randint(1, 1000),
            "IMAGES_THUMBS": {
                'small': (random.randint(1, 1000), random.randint(1, 1000)),
                'big': (random.randint(1, 1000), random.randint(1, 1000))
            }
        }
        if not prefix:
            return settings

        return {prefix.upper() + "_" + k if k != "IMAGES_STORE" else k: v for k, v in settings.items()}

    def _generate_fake_pipeline_subclass(self):
        """
        :return: ImagePipeline class will all uppercase attributes set.
        """
        class UserDefinedImagePipeline(ImagesPipeline):
            # Values should be in different range than fake_settings.
            MIN_WIDTH = random.randint(1000, 2000)
            MIN_HEIGHT = random.randint(1000, 2000)
            THUMBS = {
                'small': (random.randint(1000, 2000), random.randint(1000, 2000)),
                'big': (random.randint(1000, 2000), random.randint(1000, 2000))
            }
            EXPIRES = random.randint(1000, 2000)
            IMAGES_URLS_FIELD = "field_one"
            IMAGES_RESULT_FIELD = "field_two"

        return UserDefinedImagePipeline

    def test_different_settings_for_different_instances(self):
        """
        If there are two instances of ImagesPipeline class with different settings, they should
        have different settings.
        """
        custom_settings = self._generate_fake_settings()
        default_settings = Settings()
        default_sts_pipe = ImagesPipeline(self.tempdir, settings=default_settings)
        user_sts_pipe = ImagesPipeline.from_settings(Settings(custom_settings))
        for pipe_attr, settings_attr in self.img_cls_attribute_names:
            expected_default_value = self.default_pipeline_settings.get(pipe_attr)
            custom_value = custom_settings.get(settings_attr)
            self.assertNotEqual(expected_default_value, custom_value)
            self.assertEqual(getattr(default_sts_pipe, pipe_attr.lower()), expected_default_value)
            self.assertEqual(getattr(user_sts_pipe, pipe_attr.lower()), custom_value)

    def test_subclass_attrs_preserved_default_settings(self):
        """
        If image settings are not defined at all subclass of ImagePipeline takes values
        from class attributes.
        """
        pipeline_cls = self._generate_fake_pipeline_subclass()
        pipeline = pipeline_cls.from_settings(Settings({"IMAGES_STORE": self.tempdir}))
        for pipe_attr, settings_attr in self.img_cls_attribute_names:
            # Instance attribute (lowercase) must be equal to class attribute (uppercase).
            attr_value = getattr(pipeline, pipe_attr.lower())
            self.assertNotEqual(attr_value, self.default_pipeline_settings[pipe_attr])
            self.assertEqual(attr_value, getattr(pipeline, pipe_attr))

    def test_subclass_attrs_preserved_custom_settings(self):
        """
        If image settings are defined but they are not defined for subclass default
        values taken from settings should be preserved.
        """
        pipeline_cls = self._generate_fake_pipeline_subclass()
        settings = self._generate_fake_settings()
        pipeline = pipeline_cls.from_settings(Settings(settings))
        for pipe_attr, settings_attr in self.img_cls_attribute_names:
            # Instance attribute (lowercase) must be equal to
            # value defined in settings.
            value = getattr(pipeline, pipe_attr.lower())
            self.assertNotEqual(value, self.default_pipeline_settings[pipe_attr])
            setings_value = settings.get(settings_attr)
            self.assertEqual(value, setings_value)

    def test_no_custom_settings_for_subclasses(self):
        """
        If there are no settings for subclass and no subclass attributes, pipeline should use
        attributes of base class.
        """
        class UserDefinedImagePipeline(ImagesPipeline):
            pass

        user_pipeline = UserDefinedImagePipeline.from_settings(Settings({"IMAGES_STORE": self.tempdir}))
        for pipe_attr, settings_attr in self.img_cls_attribute_names:
            # Values from settings for custom pipeline should be set on pipeline instance.
            custom_value = self.default_pipeline_settings.get(pipe_attr.upper())
            self.assertEqual(getattr(user_pipeline, pipe_attr.lower()), custom_value)

    def test_custom_settings_for_subclasses(self):
        """
        If there are custom settings for subclass and NO class attributes, pipeline should use custom
        settings.
        """
        class UserDefinedImagePipeline(ImagesPipeline):
            pass

        prefix = UserDefinedImagePipeline.__name__.upper()
        settings = self._generate_fake_settings(prefix=prefix)
        user_pipeline = UserDefinedImagePipeline.from_settings(Settings(settings))
        for pipe_attr, settings_attr in self.img_cls_attribute_names:
            # Values from settings for custom pipeline should be set on pipeline instance.
            custom_value = settings.get(prefix + "_" + settings_attr)
            self.assertNotEqual(custom_value, self.default_pipeline_settings[pipe_attr])
            self.assertEqual(getattr(user_pipeline, pipe_attr.lower()), custom_value)

    def test_custom_settings_and_class_attrs_for_subclasses(self):
        """
        If there are custom settings for subclass AND class attributes
        setting keys are preferred and override attributes.
        """
        pipeline_cls = self._generate_fake_pipeline_subclass()
        prefix = pipeline_cls.__name__.upper()
        settings = self._generate_fake_settings(prefix=prefix)
        user_pipeline = pipeline_cls.from_settings(Settings(settings))
        for pipe_attr, settings_attr in self.img_cls_attribute_names:
            custom_value = settings.get(prefix + "_" + settings_attr)
            self.assertNotEqual(custom_value, self.default_pipeline_settings[pipe_attr])
            self.assertEqual(getattr(user_pipeline, pipe_attr.lower()), custom_value)

    def test_cls_attrs_with_DEFAULT_prefix(self):
        class UserDefinedImagePipeline(ImagesPipeline):
            DEFAULT_IMAGES_URLS_FIELD = "something"
            DEFAULT_IMAGES_RESULT_FIELD = "something_else"
        pipeline = UserDefinedImagePipeline.from_settings(Settings({"IMAGES_STORE": self.tempdir}))
        self.assertEqual(pipeline.images_result_field, "something_else")
        self.assertEqual(pipeline.images_urls_field, "something")

    def test_user_defined_subclass_default_key_names(self):
        """Test situation when user defines subclass of ImagePipeline,
        but uses attribute names for default pipeline (without prefixing
        them with pipeline class name).
        """
        settings = self._generate_fake_settings()

        class UserPipe(ImagesPipeline):
            pass

        pipeline_cls = UserPipe.from_settings(Settings(settings))

        for pipe_attr, settings_attr in self.img_cls_attribute_names:
            expected_value = settings.get(settings_attr)
            self.assertEqual(getattr(pipeline_cls, pipe_attr.lower()),
                             expected_value)


def _create_image(format, *a, **kw):
    buf = io.BytesIO()
    Image.new(*a, **kw).save(buf, format)
    buf.seek(0)
    return Image.open(buf)


if __name__ == "__main__":
    unittest.main()



# coding=utf-8
from twisted.trial import unittest


def reppy_available():
    # check if reppy parser is installed
    try:
        from reppy.robots import Robots  # noqa: F401
    except ImportError:
        return False
    return True


def rerp_available():
    # check if robotexclusionrulesparser is installed
    try:
        from robotexclusionrulesparser import RobotExclusionRulesParser  # noqa: F401
    except ImportError:
        return False
    return True


def protego_available():
    # check if protego parser is installed
    try:
        from protego import Protego  # noqa: F401
    except ImportError:
        return False
    return True


class BaseRobotParserTest:
    def _setUp(self, parser_cls):
        self.parser_cls = parser_cls

    def test_allowed(self):
        robotstxt_robotstxt_body = ("User-agent: * \n"
                                    "Disallow: /disallowed \n"
                                    "Allow: /allowed \n"
                                    "Crawl-delay: 10".encode('utf-8'))
        rp = self.parser_cls.from_crawler(crawler=None, robotstxt_body=robotstxt_robotstxt_body)
        self.assertTrue(rp.allowed("https://www.site.local/allowed", "*"))
        self.assertFalse(rp.allowed("https://www.site.local/disallowed", "*"))

    def test_allowed_wildcards(self):
        robotstxt_robotstxt_body = """User-agent: first
                                Disallow: /disallowed/*/end$

                                User-agent: second
                                Allow: /*allowed
                                Disallow: /
                                """.encode('utf-8')
        rp = self.parser_cls.from_crawler(crawler=None, robotstxt_body=robotstxt_robotstxt_body)

        self.assertTrue(rp.allowed("https://www.site.local/disallowed", "first"))
        self.assertFalse(rp.allowed("https://www.site.local/disallowed/xyz/end", "first"))
        self.assertFalse(rp.allowed("https://www.site.local/disallowed/abc/end", "first"))
        self.assertTrue(rp.allowed("https://www.site.local/disallowed/xyz/endinglater", "first"))

        self.assertTrue(rp.allowed("https://www.site.local/allowed", "second"))
        self.assertTrue(rp.allowed("https://www.site.local/is_still_allowed", "second"))
        self.assertTrue(rp.allowed("https://www.site.local/is_allowed_too", "second"))

    def test_length_based_precedence(self):
        robotstxt_robotstxt_body = ("User-agent: * \n"
                                    "Disallow: / \n"
                                    "Allow: /page".encode('utf-8'))
        rp = self.parser_cls.from_crawler(crawler=None, robotstxt_body=robotstxt_robotstxt_body)
        self.assertTrue(rp.allowed("https://www.site.local/page", "*"))

    def test_order_based_precedence(self):
        robotstxt_robotstxt_body = ("User-agent: * \n"
                                    "Disallow: / \n"
                                    "Allow: /page".encode('utf-8'))
        rp = self.parser_cls.from_crawler(crawler=None, robotstxt_body=robotstxt_robotstxt_body)
        self.assertFalse(rp.allowed("https://www.site.local/page", "*"))

    def test_empty_response(self):
        """empty response should equal 'allow all'"""
        rp = self.parser_cls.from_crawler(crawler=None, robotstxt_body=b'')
        self.assertTrue(rp.allowed("https://site.local/", "*"))
        self.assertTrue(rp.allowed("https://site.local/", "chrome"))
        self.assertTrue(rp.allowed("https://site.local/index.html", "*"))
        self.assertTrue(rp.allowed("https://site.local/disallowed", "*"))

    def test_garbage_response(self):
        """garbage response should be discarded, equal 'allow all'"""
        robotstxt_robotstxt_body = b'GIF89a\xd3\x00\xfe\x00\xa2'
        rp = self.parser_cls.from_crawler(crawler=None, robotstxt_body=robotstxt_robotstxt_body)
        self.assertTrue(rp.allowed("https://site.local/", "*"))
        self.assertTrue(rp.allowed("https://site.local/", "chrome"))
        self.assertTrue(rp.allowed("https://site.local/index.html", "*"))
        self.assertTrue(rp.allowed("https://site.local/disallowed", "*"))

    def test_unicode_url_and_useragent(self):
        robotstxt_robotstxt_body = """
        User-Agent: *
        Disallow: /admin/
        Disallow: /static/
        # taken from https://en.wikipedia.org/robots.txt
        Disallow: /wiki/K%C3%A4ytt%C3%A4j%C3%A4:
        Disallow: /wiki/Käyttäjä:

        User-Agent: UnicödeBöt
        Disallow: /some/randome/page.html""".encode('utf-8')
        rp = self.parser_cls.from_crawler(crawler=None, robotstxt_body=robotstxt_robotstxt_body)
        self.assertTrue(rp.allowed("https://site.local/", "*"))
        self.assertFalse(rp.allowed("https://site.local/admin/", "*"))
        self.assertFalse(rp.allowed("https://site.local/static/", "*"))
        self.assertTrue(rp.allowed("https://site.local/admin/", "UnicödeBöt"))
        self.assertFalse(rp.allowed("https://site.local/wiki/K%C3%A4ytt%C3%A4j%C3%A4:", "*"))
        self.assertFalse(rp.allowed("https://site.local/wiki/Käyttäjä:", "*"))
        self.assertTrue(rp.allowed("https://site.local/some/randome/page.html", "*"))
        self.assertFalse(rp.allowed("https://site.local/some/randome/page.html", "UnicödeBöt"))


class PythonRobotParserTest(BaseRobotParserTest, unittest.TestCase):
    def setUp(self):
        from scrapy.robotstxt import PythonRobotParser
        super()._setUp(PythonRobotParser)

    def test_length_based_precedence(self):
        raise unittest.SkipTest("RobotFileParser does not support length based directives precedence.")

    def test_allowed_wildcards(self):
        raise unittest.SkipTest("RobotFileParser does not support wildcards.")


class ReppyRobotParserTest(BaseRobotParserTest, unittest.TestCase):
    if not reppy_available():
        skip = "Reppy parser is not installed"

    def setUp(self):
        from scrapy.robotstxt import ReppyRobotParser
        super()._setUp(ReppyRobotParser)

    def test_order_based_precedence(self):
        raise unittest.SkipTest("Reppy does not support order based directives precedence.")


class RerpRobotParserTest(BaseRobotParserTest, unittest.TestCase):
    if not rerp_available():
        skip = "Rerp parser is not installed"

    def setUp(self):
        from scrapy.robotstxt import RerpRobotParser
        super()._setUp(RerpRobotParser)

    def test_length_based_precedence(self):
        raise unittest.SkipTest("Rerp does not support length based directives precedence.")


class ProtegoRobotParserTest(BaseRobotParserTest, unittest.TestCase):
    if not protego_available():
        skip = "Protego parser is not installed"

    def setUp(self):
        from scrapy.robotstxt import ProtegoRobotParser
        super()._setUp(ProtegoRobotParser)

    def test_order_based_precedence(self):
        raise unittest.SkipTest("Protego does not support order based directives precedence.")



# -*- coding: utf-8 -*-

"""
原厂物料
FactoryMaterialItem
FactoryCatergoryItem
原厂资讯
FactoryNewsItem
FactoryNewsMapItem
FactoryNewsCategoryItem
原厂方案
FactoryPlanItem
FactoryCategoryItem
FactoryPlanMapItem
原厂课程视频
FactoryVideoItem
FactoryCourseItem
FactoryCourseCategoryItem
FactoryCourseMapItem
代理商
FactoryAgentItem
原厂pcnpdn
PCNPDNItem
原厂百科
FactoryBaikeItem
原厂价格
EbusinessPrice
"""

# Define here the models for your scraped items
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/items.html
import scrapy


# 重要字段用'<----'标识,如果与需求不符合要找需求人对一下

# 原厂物料
class FactoryMaterialItem(scrapy.Item):
    # table = 'ware_detail'  # 测试库bd-crawler 正式库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryMaterialItem, self).__init__()

    table = 'ware_detail'
    sources = scrapy.Field()  # 来源：oneyac-唯样，szlcsc-立创，datasheet5
    url = scrapy.Field()  # 详情链接 <----
    title = scrapy.Field()  # 型号  <----
    img_url = scrapy.Field()  # 图片链接,如果有多个会以 | 隔开  <----
    raw_img_url = scrapy.Field()  # 图片链接,如果有多个会以 | 隔开  <----
    pdf_url = scrapy.Field()  # pdf链接,如果有多个会以 | 隔开  <----
    raw_pdf_url = scrapy.Field()  # pdf链接,如果有多个会以 | 隔开  <----
    category_name = scrapy.Field()  # 分类名称  <----
    category_id = scrapy.Field()  # xcc分类id  <---- {原厂英文名}_{分类英文名或者拼音(不要留空格)}
    brand_name = scrapy.Field()  # 品牌名称
    brand_id = scrapy.Field()  # xcc品牌id
    descs = scrapy.Field()  # 描述 <----
    container_json = scrapy.Field()  # 头的全部内容
    other_url_json = scrapy.Field()  # 其他链接的json
    list_json = scrapy.Field()  # 列表json <---- 参数数据json单层结构
    number = scrapy.Field()  # 总数量
    spider_flag = scrapy.Field()  # 是否爬取当前url: 1是，0否,null否
    spider_number = scrapy.Field()  # 已爬取数量
    create_time = scrapy.Field()  # 创建时间 <----
    creator = scrapy.Field()  # 创建人 <-----
    packing = scrapy.Field()  # 封装，同型号的多个用，号隔开 <----
    price_json = scrapy.Field()  # 价格json，格式如：[{"price":"1.2","qty":1}]
    mpq = scrapy.Field()  # 最小包装量
    moq = scrapy.Field()  # 最小起订量
    min_work_tp = scrapy.Field()  # 最低工作温度(单位摄氏度)
    max_work_tp = scrapy.Field()  # 最高工作温度(单位摄氏度)
    rough_weight = scrapy.Field()  # 毛重(单位g)
    manner_packing = scrapy.Field()  # 包装方式
    layout_design = scrapy.Field()  # 电路图
    other_pdf_url = scrapy.Field()  # 其他规格书{"名称":"pdf链接"}
    raw_other_pdf_url = scrapy.Field()  # 其他规格书{"名称":"pdf链接"}
    interior_structure = scrapy.Field() # 内部结构
    installation_type=scrapy.Field()    # 安装类型
    raw_layout_design = scrapy.Field()  # 电路图原链接


class FactoryCatergoryItem(scrapy.Item):
    table = 'ware_category'  # 测试库bd-crawler 正式库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryCatergoryItem, self).__init__()

    sources = scrapy.Field()  # 来源：oneyac-唯样，szlcsc-立创<-----
    url = scrapy.Field()  # 链接<-----
    level = scrapy.Field()  # 分类级别<-----
    category_name = scrapy.Field()  # 分类名称<-----
    en_name = scrapy.Field()  # 英文-分类名称
    category_id = scrapy.Field()  # 分类id<-----
    p_category_id = scrapy.Field()  # 父类-分类id<----- 如果没有父级不用写
    p_category_name = scrapy.Field()  # 父类-分类名称<-----
    children_number = scrapy.Field()  # 子类数量
    img_url = scrapy.Field()  # 图片链接
    jsons = scrapy.Field()  # 额外参数
    attrs = scrapy.Field()  # 属性名称列表
    number = scrapy.Field()  # 总数量
    spider_flag = scrapy.Field()  # 是否爬取当前url: 1是，0否
    spider_number = scrapy.Field()  # 已爬取数量
    spider_date = scrapy.Field()  # 爬虫年月日
    spider_time = scrapy.Field()  # 爬虫时间
    create_time = scrapy.Field()  # 创建时间
    creator = scrapy.Field()  # 创建人 <-----
    update_time = scrapy.Field()  # 更新时间
    updator = scrapy.Field()  # 更新人
    del_flag = scrapy.Field()  # 删除标记
    brand_name = scrapy.Field()  # 品牌名称 <-----
    brand_id = scrapy.Field()  # xcc品牌id <-----


# 原厂资讯
class FactoryNewsItem(scrapy.Item):
    table = 'news_detail'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryNewsItem, self).__init__()

    source = scrapy.Field()  # *来源 <-----
    category = scrapy.Field()  # *分类<-----
    source_url = scrapy.Field()  # *来源链接<-----
    title = scrapy.Field()  # *标题<-----
    content = scrapy.Field()  # *内容<-----
    article_from = scrapy.Field()  # *文章来源<-----
    author = scrapy.Field()  # 作者
    tips = scrapy.Field()  # 标签
    publish_time = scrapy.Field()  # *新闻时间<-----
    create_time = scrapy.Field()  # *创建时间<-----
    creator = scrapy.Field()  # *创建人<-----
    update_time = scrapy.Field()  # 更新时间
    updator = scrapy.Field()  # 更新人
    spider_name = scrapy.Field()  # 爬虫名
    spider_time = scrapy.Field()  # 爬虫时间
    del_flag = scrapy.Field()  # 删除标记
    note = scrapy.Field()  # 备注
    img_url = scrapy.Field()  # *图片链接(文章图片链接"|"做分割) <-----
    img_num = scrapy.Field()  # 用做图片数量(弃用不用写)
    new3 = scrapy.Field()  # 资讯是否公开(弃用不用写)
    abstract = scrapy.Field()  # 摘要
    read_number = scrapy.Field()  # 浏览次数(文章没有则随机2000以内) <-----
    is_open = scrapy.Field()  # 资讯是否公开(网信资讯设置为0) <----- 需要确认是否是网信资讯
    title_en = scrapy.Field()  # 英文标题
    content_en = scrapy.Field()  # 英文内容
    tips_en = scrapy.Field()  # 英文标签
    is_en = scrapy.Field()  # 是否外文（0：否；1：是）(接口设置参数,不用写)
    appdetail_id = scrapy.Field()  # uuid对应品牌id(品牌资讯需要,普通资讯不用写) <----- hash.sha1模式与原厂关联


class FactoryNewsMapItem(scrapy.Item):
    table = 'app_detail_brand'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryNewsMapItem, self).__init__()

    appdetail_id = scrapy.Field()  # <-----
    brand_id = scrapy.Field()  # <-----


class FactoryNewsCategoryItem(scrapy.Item):
    table = 'news_category'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryNewsCategoryItem, self).__init__()

    sources = scrapy.Field()  # 来源：cena
    url = scrapy.Field()  # 链接
    level = scrapy.Field()  # 分类级别
    category_name = scrapy.Field()  # 分类名称
    en_name = scrapy.Field()  # 英文-分类名称
    category_id = scrapy.Field()  # 分类id
    p_category_id = scrapy.Field()  # 父类-分类id
    p_category_name = scrapy.Field()  # 父类-分类名称
    p_number = scrapy.Field()  # 子类数量
    jsons = scrapy.Field()  # 额外参数
    attrs = scrapy.Field()  # 属性名称列表
    number = scrapy.Field()  # 总数量
    spider_flag = scrapy.Field()  # 是否爬取当前url: 1是，0否
    spider_number = scrapy.Field()  # 已爬取数量
    spider_date = scrapy.Field()  # 爬虫年月日
    spider_time = scrapy.Field()  # 爬虫时间
    create_time = scrapy.Field()  # 创建时间
    creator = scrapy.Field()  # 创建人
    update_time = scrapy.Field()  # 更新时间
    updator = scrapy.Field()  # 更新人
    del_flag = scrapy.Field()  # 删除标记


# 原厂方案
class FactoryPlanItem(scrapy.Item):
    table = 'plan_info'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryPlanItem, self).__init__()

    name = scrapy.Field()  # 名称 相当于资讯title  <-----
    style = scrapy.Field()  # 分类 相当于资讯 sources <-----
    type_first = scrapy.Field()  # 一级应用分类 相当于资讯 category_id <-----
    type_second = scrapy.Field()  # 二级应用分类 相当于资讯 category_id <-----
    des = scrapy.Field()  # 简介 列表页简介不可带html标签
    detail_des = scrapy.Field()  # 详细描述 相当于资讯的contain 正文内容 <-----
    image_url = scrapy.Field()  # 图片资源 资源图片多张用|分割 <-----
    status = scrapy.Field()  # 审批状态
    time = scrapy.Field()  # 发布时间 相当于资讯publish_time  <-----
    create_by_user = scrapy.Field()  # 创建人  <-----
    create_datetime = scrapy.Field()  # 创建时间 <-----
    scan_num = scrapy.Field()  # 浏览次数  <-----
    collect_num = scrapy.Field()  # 收藏次数
    pdf_path = scrapy.Field()  # pdf路径(设计下载)
    related_documents_pdf = scrapy.Field()  # 相关文档
    plan_url = scrapy.Field()  # 方案链接
    tips = scrapy.Field()  # 标签 <-----
    no = scrapy.Field()  # 编号，用于爬虫除重
    plan_id = scrapy.Field()  # uuid,  用于关联品牌 <-----


class FactoryCategoryItem(scrapy.Item):
    table = 'plan_category'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryCategoryItem, self).__init__()

    name = scrapy.Field()  # 一级分类名称 相当于资讯  <----- 展示字段
    type_first = scrapy.Field()  # 一级分类标识 相当于资讯 category_id <----- 关联字段
    second_name = scrapy.Field()  # 二级分类名称 相当于资讯  <----- 展示字段
    type_second = scrapy.Field()  # 二级分类标识 相当于资讯 category_id <----- 关联字段
    style = scrapy.Field()  # 分类 相当于资讯 sources <-----


class FactoryPlanMapItem(scrapy.Item):
    table = 'app_plan_brand'

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryPlanMapItem, self).__init__()

    plan_id = scrapy.Field()  # uuid,  用于关联品牌 <-----
    brand_id = scrapy.Field()  # 原厂id <-----


# 原厂课程视频
class FactoryVideoItem(scrapy.Item):
    table = 'app_video'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryVideoItem, self).__init__()

    title = scrapy.Field()  # 标题 <-----
    source = scrapy.Field()  # 来源 <-----
    publishtime = scrapy.Field()  # 发布时间 <-----
    link_address = scrapy.Field()  # 本地链接地址 <-----
    is_checked = scrapy.Field()  # 是否审核，默认0未审核，1 审核
    create_time = scrapy.Field()  # 创建时间 <-----
    creator = scrapy.Field()  # 创建人 <-----
    is_deleted = scrapy.Field()  # 是否删除 0删除，1 不删除
    course_url = scrapy.Field()  # 课程url,映射父级用处 <----- 关联字段
    original_url = scrapy.Field()  # 原链接 <-----
    sort = scrapy.Field()  # 排序 视频课程的序列号,如:课时二 写2 (int类型) <-----


class FactoryCourseItem(scrapy.Item):
    table = 'app_course'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryCourseItem, self).__init__()

    category_id = scrapy.Field()  # 分类ID
    course_name = scrapy.Field()  # 课程名
    snapshot = scrapy.Field()  # 缩略图 <-----课程图片,详情页中取或者列表页取
    introduction = scrapy.Field()  # 介绍 <-----课程介绍
    source = scrapy.Field()  # 来源 <-----
    publishtime = scrapy.Field()  # 发布日期 <-----
    is_checked = scrapy.Field()  # 是否审核，默认0未审核，1已审核
    create_time = scrapy.Field()  # 创建时间 <-----
    creator = scrapy.Field()  # 创建人 <-----
    tips = scrapy.Field()  # 标签 <-----
    author_intro = scrapy.Field()  # 作者简介 <-----
    course_url = scrapy.Field()  # 课程url,映射video作用 <----- 详情表关联字段
    category_url = scrapy.Field()  # 分类url,映射课程分类作用 <----- 父级关联字段
    course_id = scrapy.Field()  # 课程UUID <-----hash.sha1模式 用于关联品牌


class FactoryCourseCategoryItem(scrapy.Item):
    table = 'app_course_category'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryCourseCategoryItem, self).__init__()

    category_name = scrapy.Field()  # 分类名称 <-----
    create_time = scrapy.Field()  # 创建时间 <-----
    creator = scrapy.Field()  # 创建人 <-----
    category_url = scrapy.Field()  # 分类url,映射课程作用 <-----
    sort = scrapy.Field()  # 排序 <-----


class FactoryCourseMapItem(scrapy.Item):
    table = 'app_course_brand'  # 正式库/测试库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryCourseMapItem, self).__init__()

    course_id = scrapy.Field()  # uuid,  用于关联品牌 <-----
    brand_name = scrapy.Field()  # 原厂id <-----


# 代理商
class FactoryAgentItem(scrapy.Item):
    table = 'app_agent_copy1'

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryAgentItem, self).__init__()

    brand_id = scrapy.Field()  # 品牌ID
    brand_name_cn = scrapy.Field()  # 品牌中文名称
    brand_name_en = scrapy.Field()  # 品牌英文名称
    group_name = scrapy.Field()  # 分组名称
    agent_name = scrapy.Field()  # 代理商中文名称
    email = scrapy.Field()  # 邮箱
    phone = scrapy.Field()  # 电话号码
    country = scrapy.Field()  # 国家
    city = scrapy.Field()  # 城市
    address = scrapy.Field()  # 地址
    create_time = scrapy.Field()  # 创建时间


# PCN/PDN
class PCNPDNItem(scrapy.Item):
    table = 'pcn_pdn_new'

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(PCNPDNItem, self).__init__()

    sources = scrapy.Field()  # 来源,原厂Manufacturers
    type = scrapy.Field()  # 必填，格式PCN或PDN <----
    title = scrapy.Field()  # 必填，型号<----
    brand_name = scrapy.Field()  # 必填，品牌名<----
    pdf_file = scrapy.Field()  # 必填，PCN/PDN链接文件<----
    pdf_url = scrapy.Field()  # 必填,pdf链接
    brand_id = scrapy.Field()  # 必填,品牌id
    creator = scrapy.Field()  # 创建人
    create_time = scrapy.Field()  # 创建时间
    update_time = scrapy.Field()  # 更新时间
    change_content = scrapy.Field()  # 变更内容
    no = scrapy.Field()  # 编码


class FactoryBaikeItem(scrapy.Item):
    table = 'ware_wiki'

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(FactoryBaikeItem, self).__init__()

    sources = scrapy.Field()  # 来源：oneyac-唯样，szlcsc-立创，datasheet5
    url = scrapy.Field()  # 详情链接 <----
    title = scrapy.Field()  # 型号  <----
    brand_name = scrapy.Field()
    brand_id = scrapy.Field()  # xcc品牌id <-----
    category_name = scrapy.Field()  # 分类名称  <----
    category_id = scrapy.Field()  # xcc分类id  <---- {原厂英文名}_{分类英文名或者拼音(不要留空格)}
    create_time = scrapy.Field()  # 创建时间 <----
    creator = scrapy.Field()  # 创建人 <-----
    application = scrapy.Field()  # 应用 <-----
    feature = scrapy.Field()  # 特点特性 <-----
    standard = scrapy.Field()  # 规格 <-----
    description = scrapy.Field()  # 描述 <-----
    overview = scrapy.Field()  # 概述 <-----
    spider_date = scrapy.Field()  # 概述 <-----


class EbusinessPrice(scrapy.Item):
    table = 'ware_detail_price'  # 测试库bd-crawler 正式库bd-spider

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(EbusinessPrice, self).__init__()

    sources = scrapy.Field()  # 来源：oneyac-唯样，szlcsc-立创，datasheet5
    url = scrapy.Field()  # 详情链接 <----
    title = scrapy.Field()  # 型号  <----
    unit = scrapy.Field()  # CNY(人民币)标识为0 USD(美元)标识为1 <----
    price_json = scrapy.Field()  # 价格
    brand_id = scrapy.Field()  # 品牌ID
    stock = scrapy.Field()  # 库存数量  <----
    in_stock = scrapy.Field()  # 是否有货  0无货 1有货
    brand_name = scrapy.Field()  # <----厂商


class officeItem(scrapy.Item):
    table = 'app_office'

    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(officeItem, self).__init__()

    sources = scrapy.Field()  # <---数据来源,取网站域名
    brand_name = scrapy.Field()  # 品牌名称
    name = scrapy.Field()  # <---办事处名称
    mobile = scrapy.Field()  # <--- 办事处电话
    address = scrapy.Field()  # <--- 办事处地址
    agency_id = scrapy.Field()  # <--- 代理商id
    email = scrapy.Field()  # <--- 办事处地址
    city = scrapy.Field()  # <--- 城市信息
    country = scrapy.Field()  # <--- 国家
    create_time = scrapy.Field()  # <--- 创建时间


# 数据平台
class DetailDataplatformItem(scrapy.Item):
    table = 'ware_detail_dataplatform'
    sources = scrapy.Field()  # 来源：oneyac-唯样，szlcsc-立创，datasheet5
    url = scrapy.Field()  # 详情链接 <----
    title = scrapy.Field()  # 型号  <----
    brand_name = scrapy.Field()
    brand_id = scrapy.Field()  # 品牌ID
    descs = scrapy.Field()
    max_work_tp = scrapy.Field()
    min_work_tp = scrapy.Field()
    pdf_url = scrapy.Field()
    raw_pdf_url = scrapy.Field()
    img_url = scrapy.Field()
    raw_img_url = scrapy.Field()
    container_json = scrapy.Field()
    list_json = scrapy.Field()
    creator = scrapy.Field()
    create_time = scrapy.Field()
    packing = scrapy.Field()
    id = scrapy.Field()
    category_name = scrapy.Field()
    category_id = scrapy.Field()
    layout_design = scrapy.Field()
    other_pdf_url = scrapy.Field()
    spider_date = scrapy.Field()
    raw_id = scrapy.Field()


# 热搜
class DetailResouItem(scrapy.Item):
    table = 'ware_resou'
    sources = scrapy.Field()
    url = scrapy.Field()
    title = scrapy.Field()
    brand_name = scrapy.Field()
    brand_id = scrapy.Field()
    descs = scrapy.Field()
    max_work_tp = scrapy.Field()
    min_work_tp = scrapy.Field()
    pdf_url = scrapy.Field()
    raw_pdf_url = scrapy.Field()
    img_url = scrapy.Field()
    raw_img_url = scrapy.Field()
    list_json = scrapy.Field()
    creator = scrapy.Field()
    create_time = scrapy.Field()
    packing = scrapy.Field()
    id = scrapy.Field()
    category_name = scrapy.Field()
    category_id = scrapy.Field()
    layout_design = scrapy.Field()
    other_pdf_url = scrapy.Field()
    spider_time = scrapy.Field()


# pip to pin
class WarePinformItem(scrapy.Item):
    table = 'ware_pin'
    sources = scrapy.Field()
    url = scrapy.Field()
    level = scrapy.Field()
    title = scrapy.Field()
    brand_name = scrapy.Field()
    brand_id = scrapy.Field()
    pin_url = scrapy.Field()
    pin_title = scrapy.Field()
    pin_brand_id = scrapy.Field()
    pin_brand_name = scrapy.Field()
    creator = scrapy.Field()
    create_time = scrapy.Field()
    update_time = scrapy.Field()
    comments = scrapy.Field()
    category_name = scrapy.Field()
    category_id = scrapy.Field()
    image_url = scrapy.Field()
    raw_image_url = scrapy.Field()
    list_json = scrapy.Field()
    pin_list_json = scrapy.Field()
    pin_image_url = scrapy.Field()
    pin_raw_image_url = scrapy.Field()
    description = scrapy.Field()


class DetailPinformItem(scrapy.Item):
    table = 'ware_agg_pin'
    sources = scrapy.Field()
    url = scrapy.Field()
    category_name = scrapy.Field()
    category_id = scrapy.Field()
    title = scrapy.Field()
    brand_name = scrapy.Field()
    brand_id = scrapy.Field()
    creator = scrapy.Field()
    create_time = scrapy.Field()
    level = scrapy.Field()
    similarity = scrapy.Field()
    similarity_json = scrapy.Field()
    pin_similarity_json = scrapy.Field()
    image_url = scrapy.Field()
    raw_image_url = scrapy.Field()
    image_url_name = scrapy.Field()
    list_json = scrapy.Field()
    pin_list_json = scrapy.Field()
    function_json = scrapy.Field()
    pin_function_json = scrapy.Field()
    packing_json = scrapy.Field()
    pin_packing_json = scrapy.Field()
    plan_json = scrapy.Field()
    pin_plan_json = scrapy.Field()
    pin_title = scrapy.Field()
    pin_brand_id = scrapy.Field()
    pin_brand_name = scrapy.Field()
    pin_image_url = scrapy.Field()
    pin_raw_image_url = scrapy.Field()
    pin_image_url_name = scrapy.Field()
    pin_url = scrapy.Field()
    update_time = scrapy.Field()
    comments = scrapy.Field()


class SmdITEM(scrapy.Item):
    # table = 'ware_detail'  # 测试库bd-crawler 正式库bd-spider
    # lyh丝印
    def __init__(self, *args, **kwargs):
        table_params = {'table': kwargs.get('table', self.table)} if not kwargs.get('table_add', '') \
            else {'table': kwargs.get('table', self.table) + kwargs.get('table_add', '')}
        self.__dict__.update(table_params)
        self._values = {}
        if args or kwargs and 'table' not in kwargs and 'table_add' not in kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        super(SmdITEM, self).__init__()

    table = 'ware_smd_lyh'  # 正式库/测试库bd-spider
    smd = scrapy.Field()
    title = scrapy.Field()
    brand_id = scrapy.Field()
    brand_name = scrapy.Field()
    packages = scrapy.Field()
    update_time = scrapy.Field()
    count = scrapy.Field()
    search_word = scrapy.Field()
    describee = scrapy.Field()
    sources = scrapy.Field()
    bak1 = scrapy.Field()
    sources_url = scrapy.Field()
    create_user = scrapy.Field()
    status = scrapy.Field()



from testfixtures import LogCapture
from twisted.internet import defer

from scrapy.exceptions import StopDownload

from tests.test_engine import (
    AttrsItemsSpider,
    DataClassItemsSpider,
    DictItemsSpider,
    TestSpider,
    CrawlerRun,
    EngineTest,
)


class BytesReceivedCrawlerRun(CrawlerRun):
    def bytes_received(self, data, request, spider):
        super().bytes_received(data, request, spider)
        raise StopDownload(fail=False)


class BytesReceivedEngineTest(EngineTest):
    @defer.inlineCallbacks
    def test_crawler(self):
        for spider in (TestSpider, DictItemsSpider, AttrsItemsSpider, DataClassItemsSpider):
            if spider is None:
                continue
            self.run = BytesReceivedCrawlerRun(spider)
            with LogCapture() as log:
                yield self.run.run()
                log.check_present(("scrapy.core.downloader.handlers.http11",
                                   "DEBUG",
                                   f"Download stopped for <GET http://localhost:{self.run.portno}/redirected> "
                                   "from signal handler BytesReceivedCrawlerRun.bytes_received"))
                log.check_present(("scrapy.core.downloader.handlers.http11",
                                   "DEBUG",
                                   f"Download stopped for <GET http://localhost:{self.run.portno}/> "
                                   "from signal handler BytesReceivedCrawlerRun.bytes_received"))
                log.check_present(("scrapy.core.downloader.handlers.http11",
                                   "DEBUG",
                                   f"Download stopped for <GET http://localhost:{self.run.portno}/numbers> "
                                   "from signal handler BytesReceivedCrawlerRun.bytes_received"))
            self._assert_visited_urls()
            self._assert_scheduled_requests(count=9)
            self._assert_downloaded_responses(count=9)
            self._assert_signals_caught()
            self._assert_headers_received()
            self._assert_bytes_received()

    def _assert_bytes_received(self):
        self.assertEqual(9, len(self.run.bytes))
        for request, data in self.run.bytes.items():
            joined_data = b"".join(data)
            self.assertTrue(len(data) == 1)  # signal was fired only once
            if self.run.getpath(request.url) == "/numbers":
                # Received bytes are not the complete response. The exact amount depends
                # on the buffer size, which can vary, so we only check that the amount
                # of received bytes is strictly less than the full response.
                numbers = [str(x).encode("utf8") for x in range(2**18)]
                self.assertTrue(len(joined_data) < len(b"".join(numbers)))



"""
Scrapy engine tests

This starts a testing web server (using twisted.server.Site) and then crawls it
with the Scrapy crawler.

To view the testing web server in a browser you can start it by running this
module with the ``runserver`` argument::

    python test_engine.py runserver
"""

import os
import re
import sys
import warnings
from collections import defaultdict
from urllib.parse import urlparse

import attr
from itemadapter import ItemAdapter
from pydispatch import dispatcher
from twisted.internet import defer, reactor
from twisted.trial import unittest
from twisted.web import server, static, util

from scrapy import signals
from scrapy.core.engine import ExecutionEngine
from scrapy.exceptions import CloseSpider, ScrapyDeprecationWarning
from scrapy.http import Request
from scrapy.item import Item, Field
from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import Spider
from scrapy.utils.signal import disconnect_all
from scrapy.utils.test import get_crawler

from tests import get_testdata, tests_datadir


class TestItem(Item):
    name = Field()
    url = Field()
    price = Field()


@attr.s
class AttrsItem:
    name = attr.ib(default="")
    url = attr.ib(default="")
    price = attr.ib(default=0)


class TestSpider(Spider):
    name = "scrapytest.org"
    allowed_domains = ["scrapytest.org", "localhost"]

    itemurl_re = re.compile(r"item\d+.html")
    name_re = re.compile(r"<h1>(.*?)</h1>", re.M)
    price_re = re.compile(r">Price: \$(.*?)<", re.M)

    item_cls: type = TestItem

    def parse(self, response):
        xlink = LinkExtractor()
        itemre = re.compile(self.itemurl_re)
        for link in xlink.extract_links(response):
            if itemre.search(link.url):
                yield Request(url=link.url, callback=self.parse_item)

    def parse_item(self, response):
        adapter = ItemAdapter(self.item_cls())
        m = self.name_re.search(response.text)
        if m:
            adapter['name'] = m.group(1)
        adapter['url'] = response.url
        m = self.price_re.search(response.text)
        if m:
            adapter['price'] = m.group(1)
        return adapter.item


class TestDupeFilterSpider(TestSpider):
    def start_requests(self):
        return (Request(url) for url in self.start_urls)  # no dont_filter=True


class DictItemsSpider(TestSpider):
    item_cls = dict


class AttrsItemsSpider(TestSpider):
    item_cls = AttrsItem


try:
    from dataclasses import make_dataclass
except ImportError:
    DataClassItemsSpider = None
else:
    TestDataClass = make_dataclass("TestDataClass", [("name", str), ("url", str), ("price", int)])

    class DataClassItemsSpider(DictItemsSpider):  # type: ignore[no-redef]
        def parse_item(self, response):
            item = super().parse_item(response)
            return TestDataClass(**item)


class ItemZeroDivisionErrorSpider(TestSpider):
    custom_settings = {
        "ITEM_PIPELINES": {
            "tests.pipelines.ProcessWithZeroDivisionErrorPipiline": 300,
        }
    }


class ChangeCloseReasonSpider(TestSpider):
    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        crawler.signals.connect(spider.spider_idle, signals.spider_idle)
        return spider

    def spider_idle(self):
        raise CloseSpider(reason="custom_reason")


def start_test_site(debug=False):
    root_dir = os.path.join(tests_datadir, "test_site")
    r = static.File(root_dir)
    r.putChild(b"redirect", util.Redirect(b"/redirected"))
    r.putChild(b"redirected", static.Data(b"Redirected here", "text/plain"))
    numbers = [str(x).encode("utf8") for x in range(2**18)]
    r.putChild(b"numbers", static.Data(b"".join(numbers), "text/plain"))

    port = reactor.listenTCP(0, server.Site(r), interface="127.0.0.1")
    if debug:
        print(f"Test server running at http://localhost:{port.getHost().port}/ "
              "- hit Ctrl-C to finish.")
    return port


class CrawlerRun:
    """A class to run the crawler and keep track of events occurred"""

    def __init__(self, spider_class):
        self.spider = None
        self.respplug = []
        self.reqplug = []
        self.reqdropped = []
        self.reqreached = []
        self.itemerror = []
        self.itemresp = []
        self.headers = {}
        self.bytes = defaultdict(list)
        self.signals_caught = {}
        self.spider_class = spider_class

    def run(self):
        self.port = start_test_site()
        self.portno = self.port.getHost().port

        start_urls = [
            self.geturl("/"),
            self.geturl("/redirect"),
            self.geturl("/redirect"),  # duplicate
            self.geturl("/numbers"),
        ]

        for name, signal in vars(signals).items():
            if not name.startswith('_'):
                dispatcher.connect(self.record_signal, signal)

        self.crawler = get_crawler(self.spider_class)
        self.crawler.signals.connect(self.item_scraped, signals.item_scraped)
        self.crawler.signals.connect(self.item_error, signals.item_error)
        self.crawler.signals.connect(self.headers_received, signals.headers_received)
        self.crawler.signals.connect(self.bytes_received, signals.bytes_received)
        self.crawler.signals.connect(self.request_scheduled, signals.request_scheduled)
        self.crawler.signals.connect(self.request_dropped, signals.request_dropped)
        self.crawler.signals.connect(self.request_reached, signals.request_reached_downloader)
        self.crawler.signals.connect(self.response_downloaded, signals.response_downloaded)
        self.crawler.crawl(start_urls=start_urls)
        self.spider = self.crawler.spider

        self.deferred = defer.Deferred()
        dispatcher.connect(self.stop, signals.engine_stopped)
        return self.deferred

    def stop(self):
        self.port.stopListening()
        for name, signal in vars(signals).items():
            if not name.startswith('_'):
                disconnect_all(signal)
        self.deferred.callback(None)
        return self.crawler.stop()

    def geturl(self, path):
        return f"http://localhost:{self.portno}{path}"

    def getpath(self, url):
        u = urlparse(url)
        return u.path

    def item_error(self, item, response, spider, failure):
        self.itemerror.append((item, response, spider, failure))

    def item_scraped(self, item, spider, response):
        self.itemresp.append((item, response))

    def headers_received(self, headers, body_length, request, spider):
        self.headers[request] = headers

    def bytes_received(self, data, request, spider):
        self.bytes[request].append(data)

    def request_scheduled(self, request, spider):
        self.reqplug.append((request, spider))

    def request_reached(self, request, spider):
        self.reqreached.append((request, spider))

    def request_dropped(self, request, spider):
        self.reqdropped.append((request, spider))

    def response_downloaded(self, response, spider):
        self.respplug.append((response, spider))

    def record_signal(self, *args, **kwargs):
        """Record a signal and its parameters"""
        signalargs = kwargs.copy()
        sig = signalargs.pop('signal')
        signalargs.pop('sender', None)
        self.signals_caught[sig] = signalargs


class EngineTest(unittest.TestCase):
    @defer.inlineCallbacks
    def test_crawler(self):

        for spider in (TestSpider, DictItemsSpider, AttrsItemsSpider, DataClassItemsSpider):
            if spider is None:
                continue
            self.run = CrawlerRun(spider)
            yield self.run.run()
            self._assert_visited_urls()
            self._assert_scheduled_requests(count=9)
            self._assert_downloaded_responses(count=9)
            self._assert_scraped_items()
            self._assert_signals_caught()
            self._assert_bytes_received()

    @defer.inlineCallbacks
    def test_crawler_dupefilter(self):
        self.run = CrawlerRun(TestDupeFilterSpider)
        yield self.run.run()
        self._assert_scheduled_requests(count=8)
        self._assert_dropped_requests()

    @defer.inlineCallbacks
    def test_crawler_itemerror(self):
        self.run = CrawlerRun(ItemZeroDivisionErrorSpider)
        yield self.run.run()
        self._assert_items_error()

    @defer.inlineCallbacks
    def test_crawler_change_close_reason_on_idle(self):
        self.run = CrawlerRun(ChangeCloseReasonSpider)
        yield self.run.run()
        self.assertEqual({'spider': self.run.spider, 'reason': 'custom_reason'},
                         self.run.signals_caught[signals.spider_closed])

    def _assert_visited_urls(self):
        must_be_visited = ["/", "/redirect", "/redirected",
                           "/item1.html", "/item2.html", "/item999.html"]
        urls_visited = {rp[0].url for rp in self.run.respplug}
        urls_expected = {self.run.geturl(p) for p in must_be_visited}
        assert urls_expected <= urls_visited, f"URLs not visited: {list(urls_expected - urls_visited)}"

    def _assert_scheduled_requests(self, count=None):
        self.assertEqual(count, len(self.run.reqplug))

        paths_expected = ['/item999.html', '/item2.html', '/item1.html']

        urls_requested = {rq[0].url for rq in self.run.reqplug}
        urls_expected = {self.run.geturl(p) for p in paths_expected}
        assert urls_expected <= urls_requested
        scheduled_requests_count = len(self.run.reqplug)
        dropped_requests_count = len(self.run.reqdropped)
        responses_count = len(self.run.respplug)
        self.assertEqual(scheduled_requests_count,
                         dropped_requests_count + responses_count)
        self.assertEqual(len(self.run.reqreached),
                         responses_count)

    def _assert_dropped_requests(self):
        self.assertEqual(len(self.run.reqdropped), 1)

    def _assert_downloaded_responses(self, count):
        # response tests
        self.assertEqual(count, len(self.run.respplug))
        self.assertEqual(count, len(self.run.reqreached))

        for response, _ in self.run.respplug:
            if self.run.getpath(response.url) == '/item999.html':
                self.assertEqual(404, response.status)
            if self.run.getpath(response.url) == '/redirect':
                self.assertEqual(302, response.status)

    def _assert_items_error(self):
        self.assertEqual(2, len(self.run.itemerror))
        for item, response, spider, failure in self.run.itemerror:
            self.assertEqual(failure.value.__class__, ZeroDivisionError)
            self.assertEqual(spider, self.run.spider)

            self.assertEqual(item['url'], response.url)
            if 'item1.html' in item['url']:
                self.assertEqual('Item 1 name', item['name'])
                self.assertEqual('100', item['price'])
            if 'item2.html' in item['url']:
                self.assertEqual('Item 2 name', item['name'])
                self.assertEqual('200', item['price'])

    def _assert_scraped_items(self):
        self.assertEqual(2, len(self.run.itemresp))
        for item, response in self.run.itemresp:
            item = ItemAdapter(item)
            self.assertEqual(item['url'], response.url)
            if 'item1.html' in item['url']:
                self.assertEqual('Item 1 name', item['name'])
                self.assertEqual('100', item['price'])
            if 'item2.html' in item['url']:
                self.assertEqual('Item 2 name', item['name'])
                self.assertEqual('200', item['price'])

    def _assert_headers_received(self):
        for headers in self.run.headers.values():
            self.assertIn(b"Server", headers)
            self.assertIn(b"TwistedWeb", headers[b"Server"])
            self.assertIn(b"Date", headers)
            self.assertIn(b"Content-Type", headers)

    def _assert_bytes_received(self):
        self.assertEqual(9, len(self.run.bytes))
        for request, data in self.run.bytes.items():
            joined_data = b"".join(data)
            if self.run.getpath(request.url) == "/":
                self.assertEqual(joined_data, get_testdata("test_site", "index.html"))
            elif self.run.getpath(request.url) == "/item1.html":
                self.assertEqual(joined_data, get_testdata("test_site", "item1.html"))
            elif self.run.getpath(request.url) == "/item2.html":
                self.assertEqual(joined_data, get_testdata("test_site", "item2.html"))
            elif self.run.getpath(request.url) == "/redirected":
                self.assertEqual(joined_data, b"Redirected here")
            elif self.run.getpath(request.url) == '/redirect':
                self.assertEqual(
                    joined_data,
                    b"\n<html>\n"
                    b"    <head>\n"
                    b"        <meta http-equiv=\"refresh\" content=\"0;URL=/redirected\">\n"
                    b"    </head>\n"
                    b"    <body bgcolor=\"#FFFFFF\" text=\"#000000\">\n"
                    b"    <a href=\"/redirected\">click here</a>\n"
                    b"    </body>\n"
                    b"</html>\n"
                )
            elif self.run.getpath(request.url) == "/tem999.html":
                self.assertEqual(
                    joined_data,
                    b"\n<html>\n"
                    b"  <head><title>404 - No Such Resource</title></head>\n"
                    b"  <body>\n"
                    b"    <h1>No Such Resource</h1>\n"
                    b"    <p>File not found.</p>\n"
                    b"  </body>\n"
                    b"</html>\n"
                )
            elif self.run.getpath(request.url) == "/numbers":
                # signal was fired multiple times
                self.assertTrue(len(data) > 1)
                # bytes were received in order
                numbers = [str(x).encode("utf8") for x in range(2**18)]
                self.assertEqual(joined_data, b"".join(numbers))

    def _assert_signals_caught(self):
        assert signals.engine_started in self.run.signals_caught
        assert signals.engine_stopped in self.run.signals_caught
        assert signals.spider_opened in self.run.signals_caught
        assert signals.spider_idle in self.run.signals_caught
        assert signals.spider_closed in self.run.signals_caught
        assert signals.headers_received in self.run.signals_caught

        self.assertEqual({'spider': self.run.spider},
                         self.run.signals_caught[signals.spider_opened])
        self.assertEqual({'spider': self.run.spider},
                         self.run.signals_caught[signals.spider_idle])
        self.assertEqual({'spider': self.run.spider, 'reason': 'finished'},
                         self.run.signals_caught[signals.spider_closed])

    @defer.inlineCallbacks
    def test_close_downloader(self):
        e = ExecutionEngine(get_crawler(TestSpider), lambda _: None)
        yield e.close()

    @defer.inlineCallbacks
    def test_start_already_running_exception(self):
        e = ExecutionEngine(get_crawler(TestSpider), lambda _: None)
        yield e.open_spider(TestSpider(), [])
        e.start()
        yield self.assertFailure(e.start(), RuntimeError).addBoth(
            lambda exc: self.assertEqual(str(exc), "Engine already running")
        )
        yield e.stop()

    @defer.inlineCallbacks
    def test_close_spiders_downloader(self):
        with warnings.catch_warnings(record=True) as warning_list:
            e = ExecutionEngine(get_crawler(TestSpider), lambda _: None)
            yield e.open_spider(TestSpider(), [])
            self.assertEqual(len(e.open_spiders), 1)
            yield e.close()
            self.assertEqual(len(e.open_spiders), 0)
            self.assertEqual(warning_list[0].category, ScrapyDeprecationWarning)
            self.assertEqual(
                str(warning_list[0].message),
                "ExecutionEngine.open_spiders is deprecated, please use ExecutionEngine.spider instead",
            )

    @defer.inlineCallbacks
    def test_close_engine_spiders_downloader(self):
        with warnings.catch_warnings(record=True) as warning_list:
            e = ExecutionEngine(get_crawler(TestSpider), lambda _: None)
            yield e.open_spider(TestSpider(), [])
            e.start()
            self.assertTrue(e.running)
            yield e.close()
            self.assertFalse(e.running)
            self.assertEqual(len(e.open_spiders), 0)
            self.assertEqual(warning_list[0].category, ScrapyDeprecationWarning)
            self.assertEqual(
                str(warning_list[0].message),
                "ExecutionEngine.open_spiders is deprecated, please use ExecutionEngine.spider instead",
            )

    @defer.inlineCallbacks
    def test_crawl_deprecated_spider_arg(self):
        with warnings.catch_warnings(record=True) as warning_list:
            e = ExecutionEngine(get_crawler(TestSpider), lambda _: None)
            spider = TestSpider()
            yield e.open_spider(spider, [])
            e.start()
            e.crawl(Request("data:,"), spider)
            yield e.close()
            self.assertEqual(warning_list[0].category, ScrapyDeprecationWarning)
            self.assertEqual(
                str(warning_list[0].message),
                "Passing a 'spider' argument to ExecutionEngine.crawl is deprecated",
            )

    @defer.inlineCallbacks
    def test_download_deprecated_spider_arg(self):
        with warnings.catch_warnings(record=True) as warning_list:
            e = ExecutionEngine(get_crawler(TestSpider), lambda _: None)
            spider = TestSpider()
            yield e.open_spider(spider, [])
            e.start()
            e.download(Request("data:,"), spider)
            yield e.close()
            self.assertEqual(warning_list[0].category, ScrapyDeprecationWarning)
            self.assertEqual(
                str(warning_list[0].message),
                "Passing a 'spider' argument to ExecutionEngine.download is deprecated",
            )

    @defer.inlineCallbacks
    def test_deprecated_schedule(self):
        with warnings.catch_warnings(record=True) as warning_list:
            e = ExecutionEngine(get_crawler(TestSpider), lambda _: None)
            spider = TestSpider()
            yield e.open_spider(spider, [])
            e.start()
            e.schedule(Request("data:,"), spider)
            yield e.close()
            self.assertEqual(warning_list[0].category, ScrapyDeprecationWarning)
            self.assertEqual(
                str(warning_list[0].message),
                "ExecutionEngine.schedule is deprecated, please use "
                "ExecutionEngine.crawl or ExecutionEngine.download instead",
            )

    @defer.inlineCallbacks
    def test_deprecated_has_capacity(self):
        with warnings.catch_warnings(record=True) as warning_list:
            e = ExecutionEngine(get_crawler(TestSpider), lambda _: None)
            self.assertTrue(e.has_capacity())
            spider = TestSpider()
            yield e.open_spider(spider, [])
            self.assertFalse(e.has_capacity())
            e.start()
            yield e.close()
            self.assertTrue(e.has_capacity())
            self.assertEqual(warning_list[0].category, ScrapyDeprecationWarning)
            self.assertEqual(str(warning_list[0].message), "ExecutionEngine.has_capacity is deprecated")


if __name__ == "__main__":
    if len(sys.argv) > 1 and sys.argv[1] == 'runserver':
        start_test_site(debug=True)
        reactor.run()



import unittest
from os.path import join

from w3lib.encoding import html_to_unicode

from scrapy.utils.gz import gunzip, gzip_magic_number
from scrapy.http import Response
from tests import tests_datadir


SAMPLEDIR = join(tests_datadir, 'compressed')


class GunzipTest(unittest.TestCase):

    def test_gunzip_basic(self):
        with open(join(SAMPLEDIR, 'feed-sample1.xml.gz'), 'rb') as f:
            r1 = Response("http://www.example.com", body=f.read())
            self.assertTrue(gzip_magic_number(r1))

            r2 = Response("http://www.example.com", body=gunzip(r1.body))
            self.assertFalse(gzip_magic_number(r2))
            self.assertEqual(len(r2.body), 9950)

    def test_gunzip_truncated(self):
        with open(join(SAMPLEDIR, 'truncated-crc-error.gz'), 'rb') as f:
            text = gunzip(f.read())
            assert text.endswith(b'</html')

    def test_gunzip_no_gzip_file_raises(self):
        with open(join(SAMPLEDIR, 'feed-sample1.xml'), 'rb') as f:
            self.assertRaises(IOError, gunzip, f.read())

    def test_gunzip_truncated_short(self):
        with open(join(SAMPLEDIR, 'truncated-crc-error-short.gz'), 'rb') as f:
            r1 = Response("http://www.example.com", body=f.read())
            self.assertTrue(gzip_magic_number(r1))

            r2 = Response("http://www.example.com", body=gunzip(r1.body))
            assert r2.body.endswith(b'</html>')
            self.assertFalse(gzip_magic_number(r2))

    def test_is_gzipped_empty(self):
        r1 = Response("http://www.example.com")
        self.assertFalse(gzip_magic_number(r1))

    def test_gunzip_illegal_eof(self):
        with open(join(SAMPLEDIR, 'unexpected-eof.gz'), 'rb') as f:
            text = html_to_unicode('charset=cp1252', gunzip(f.read()))[1]
            with open(join(SAMPLEDIR, 'unexpected-eof-output.txt'), 'rb') as o:
                expected_text = o.read().decode("utf-8")
                self.assertEqual(len(text), len(expected_text))
                self.assertEqual(text, expected_text)



import os,re
import shutil
import string

from importlib import import_module
from os.path import join, dirname, abspath, exists, splitext

import scrapy
from scrapy.commands import ScrapyCommand
from scrapy.utils.template import render_templatefile, string_camelcase
from scrapy.exceptions import UsageError


def sanitize_module_name(module_name):
    """Sanitize the given module name, by replacing dashes and points
    with underscores and prefixing it with a letter if it doesn't start
    with one
    """
    module_name = module_name.replace('-', '_').replace('.', '_')
    if module_name[0] not in string.ascii_letters:
        module_name = "a" + module_name
    return module_name


class Command(ScrapyCommand):

    requires_project = False
    default_settings = {'LOG_ENABLED': False}

    def syntax(self):
        return "[options] <somebody> <someproj> <name>"

    def short_desc(self):
        return "Generate new teamspider using pre-defined templates"

    def add_options(self, parser):
        ScrapyCommand.add_options(self, parser)
        parser.add_option("-l", "--list", dest="list", action="store_true",
                          help="List available templates")
        parser.add_option("-e", "--edit", dest="edit", action="store_true",
                          help="Edit spider after creating it")
        parser.add_option("-d", "--dump", dest="dump", metavar="TEMPLATE",
                          help="Dump template to standard output")
        parser.add_option("-t", "--template", dest="template", default="team",
                          help="Uses a custom template.")
        parser.add_option("--force", dest="force", action="store_true",
                          help="If the spider already exists, overwrite it with the template")

    def run(self, args, opts):
        bodydict = { # 根据成员增减,暂时写死
            "zhizhong":"zz",
            "yongjie":"yj",
            "xuqi":"xq",
            "juncheng":"jc",
            "yonghao":"yh",
            "siming":"sm",
            "fengchao":"fc",
            "runtao":"rt",
            "yijia":"yj",
            "shixi":"sx",
        }
        if opts.list:
            self._list_templates()
            return
        if opts.dump:
            template_file = self._find_template(opts.dump)
            if template_file:
                with open(template_file, "r") as f:
                    print(f.read())
            return

        # by zhizhong change project temp dir
        if len(args) != 3:
            raise UsageError()

        somebody,someproject,name = args[0:3]
        domain = 'baidu.com'
        module = sanitize_module_name(name)
        if not os.path.exists(f"./Scrapy_Prj/spiders/{somebody}"):
            os.makedirs(f"./Scrapy_Prj/spiders/{somebody}")

        if not os.path.exists(f"./Scrapy_Prj/spiders/{somebody}/__init__.py"):
            with open(f"./Scrapy_Prj/spiders/{somebody}/__init__.py","w") as f:pass

        if not os.path.exists(f"./Scrapy_Prj/spiders/{somebody}/{someproject}"):
            os.makedirs(f"./Scrapy_Prj/spiders/{somebody}/{someproject}")
            
        if not os.path.exists(f"./Scrapy_Prj/spiders/{somebody}/{someproject}/__init__.py"):
            with open(f"./Scrapy_Prj/spiders/{somebody}/{someproject}/__init__.py","w") as f:pass

        if not os.path.exists(f"./Scrapy_Prj/spiders/{somebody}/{someproject}/api.http"):
            with open(f"./Scrapy_Prj/spiders/{somebody}/{someproject}/api.http","w",encoding='utf-8') as f:f.write("#目标网站接口文档")

        if self.settings.get('BOT_NAME') == module:
            print("Cannot create a spider with the same name as your project")
            return
        spider_path =f"./Scrapy_Prj/spiders/{somebody}/{someproject}"
        if not opts.force and self._spider_exists(name):
            return
        prefix = bodydict.get(somebody,'')
        prefix_name = prefix+'_'+someproject+'_'+name if prefix else someproject+'_'+name
        self._add_crawl_register(somebody,f"{someproject}.{name}",prefix_name)

        template_file = self._find_template(opts.template)
        main_file = self._find_main_template('main')
        if template_file:
            self._genspider(module, name, domain, opts.template, template_file,spider_path)
            if opts.edit:
                self.exitcode = os.system(f'scrapy edit "{name}"')
        if main_file:
            self._genmain(module, name, domain, opts.template, main_file,somebody,someproject,bodydict)
            if opts.edit:
                self.exitcode = os.system(f'scrapy edit "{name}"')

    def _genspider(self, module, name, domain, template_name, template_file,spider_path):
        """Generate the spider module, based on the given template"""
        capitalized_module = ''.join(s.capitalize() for s in module.split('_'))
        tvars = {
            'project_name': self.settings.get('BOT_NAME'),
            'ProjectName': string_camelcase(self.settings.get('BOT_NAME')),
            'module': module,
            'name': name,
            'domain': domain,
            'classname': f'{capitalized_module}Spider'
        }
        if self.settings.get('NEWSPIDER_MODULE'):
            spiders_module = import_module(self.settings['NEWSPIDER_MODULE'])
            spiders_dir = spider_path
        else:
            spiders_module = None
            spiders_dir = "."
        spider_file = f"{join(spiders_dir, module)}.py"
        if not os.path.exists(spider_file):
            shutil.copyfile(template_file, spider_file)
            render_templatefile(spider_file, **tvars)
            print(f"Created spider {name!r} using template {template_name!r} ",
                end=('' if spiders_module else '\n'))
        else:
            print(f"Already existed {template_name!r} ",
                end=('' if spiders_module else '\n'))
        if spiders_module:
            print(f"in module:\n  {spiders_module.__name__}.{module}")

    def _genmain(self, module, name, domain, template_name, template_file,somebody,someproject,bodydict):
        """Generate the spider module, based on the given template"""
        capitalized_module = ''.join(s.capitalize() for s in module.split('_'))
        prefix = bodydict.get(somebody,'')
        prefix_name = prefix+'_'+someproject+'_'+name if prefix else someproject+'_'+name
        tvars = {
            'project_name': self.settings.get('BOT_NAME'),
            'ProjectName': string_camelcase(self.settings.get('BOT_NAME')),
            'module': module,
            'name': name,
            'prefix_name': prefix_name,
            'domain': domain,
            'classname': f'{capitalized_module}Spider'
        }
        if self.settings.get('NEWSPIDER_MODULE'):
            spiders_module = import_module(self.settings['NEWSPIDER_MODULE'])
            spiders_dir = abspath(dirname(spiders_module.__file__))
        else:
            spiders_module = None
            spiders_dir = "."

        spider_file = f"./Command/{somebody}_{someproject}_main.py"
        if not os.path.exists(spider_file):
            shutil.copyfile(template_file, spider_file)
            render_templatefile(spider_file, **tvars)
            print(f"Created {name!r} Main using template ",
                end=('' if spiders_module else '\n'))
        else:
            print(f"Already existed {name!r} Main entrance ",
                end=('\n'))
                
        if spiders_module:
            print(f"in module:\n  {spiders_module.__name__}.{module}")

    def _find_template(self, template):
        template_file = join(self.templates_dir, f'{template}.tmpl')
        if exists(template_file):
            return template_file
        print(f"Unable to find template: {template}\n")
        print('Use "scrapy genspider --list" to see all available templates.')

    def _find_main_template(self, template):
        template_file = join(self.templates_project_dir, f'{template}.tmpl')
        if exists(template_file):
            return template_file
        print(f"Unable to find template: {template}\n")
        print('Use "scrapy genspider --list" to see all available templates.')

    def _add_crawl_register(self, somebody ,spidermodule,spidername ):
        """根据命令在注册文件中添加爬虫配置
        
        Args:
            spidermodule (_type_): 爬虫模块
            spidername (_type_): 爬虫名
        """
        try: # fix spider_register.py文件写死
            with open("spider_register.py","r",encoding="utf-8") as f:
                register_text = f.read()
            join_register =  f'"{spidermodule}":["{spidername}",(),(),(),()],'
            rep_content = re.sub(somebody+r".*?\{(.*?,?)\}",somebody+r" = {\1"+"\t"+ join_register +'\n}',register_text,flags=re.S)
            with open("spider_register.py","w",encoding="utf-8") as f:
                f.write(rep_content)
        except FileNotFoundError as f:
            print("未找到spider_register注册文件")

    def _list_templates(self):
        print("Available templates:")
        for filename in sorted(os.listdir(self.templates_dir)):
            if filename.endswith('.tmpl'):
                print(f"  {splitext(filename)[0]}")

    def _spider_exists(self, name):
        if not self.settings.get('NEWSPIDER_MODULE'):
            # if run as a standalone command and file with same filename already exists
            if exists(name + ".py"):
                print(f"{abspath(name + '.py')} already exists")
                return True
            return False

        try:
            spidercls = self.crawler_process.spider_loader.load(name)
        except KeyError:
            pass
        else:
            # if spider with same name exists
            print(f"Spider {name!r} already exists in module:")
            print(f"  {spidercls.__module__}")
            return True

        # a file with the same name exists in the target directory
        spiders_module = import_module(self.settings['NEWSPIDER_MODULE'])
        spiders_dir = dirname(spiders_module.__file__)
        spiders_dir_abs = abspath(spiders_dir)
        if exists(join(spiders_dir_abs, name + ".py")):
            print(f"{join(spiders_dir_abs, (name + '.py'))} already exists")
            return True

        return False

    @property
    def templates_dir(self):
        return join(
            join(scrapy.__path__[0], 'teamtemplates'),
            'Scrapy_Prj'
        )
    @property
    def templates_project_dir(self):
        return join(scrapy.__path__[0], 'teamtemplates')



from os.path import dirname, join
import io,os
from pkg_resources import parse_version
from setuptools import setup, find_packages, __version__ as setuptools_version
import sys


with open(join(dirname(__file__), 'scrapy/VERSION'), 'rb') as f:
    version = f.read().decode('ascii').strip()

here = os.path.abspath(os.path.dirname(__file__))

magic_package_choice = 'python-magic-bin==0.4.14' if sys.platform.startswith("win") else 'python-magic==0.4.24' 

def has_environment_marker_platform_impl_support():
    """Code extracted from 'pytest/setup.py'
    https://github.com/pytest-dev/pytest/blob/7538680c/setup.py#L31

    The first known release to support environment marker with range operators
    it is 18.5, see:
    https://setuptools.readthedocs.io/en/latest/history.html#id235
    """
    return parse_version(setuptools_version) >= parse_version('18.5')

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = ''

install_requires = [
    'Twisted>=17.9.0',
    'cryptography>=2.0',
    'cssselect>=0.9.1',
    'itemloaders>=1.0.1',
    'parsel>=1.5.0',
    'pyOpenSSL>=16.2.0',
    'queuelib>=1.4.2',
    'service_identity>=16.0.0',
    'w3lib>=1.17.0',
    'zope.interface>=4.1.3',
    'protego>=0.1.15',
    'itemadapter>=0.1.0',
    'influxdb==5.3.1',
    magic_package_choice,
    "scrapyer-rabbitmq-scheduler",
    "scrapyer-redis",
]
extras_require = {}
cpython_dependencies = [
    'lxml>=3.5.0',
    'PyDispatcher>=2.0.5',
]
if has_environment_marker_platform_impl_support():
    extras_require[':platform_python_implementation == "CPython"'] = cpython_dependencies
    extras_require[':platform_python_implementation == "PyPy"'] = [
        # Earlier lxml versions are affected by
        # https://foss.heptapod.net/pypy/pypy/-/issues/2498,
        # which was fixed in Cython 0.26, released on 2017-06-19, and used to
        # generate the C headers of lxml release tarballs published since then, the
        # first of which was:
        'lxml>=4.0.0',
        'PyPyDispatcher>=2.1.0',
    ]
else:
    install_requires.extend(cpython_dependencies)


setup(
    name='Scrapy-Team',
    version=version,
    url='https://scrapy.org',
    project_urls={
        'Documentation': 'https://docs.scrapy.org/',
        'Source': 'https://github.com/buliqioqiolibusdo/scrapyer',
        'Tracker': 'https://github.com/buliqioqiolibusdo/scrapyer/issues',
    },
    description='Based on scrapy framework, combined with common component tools',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='buliqioqiolibusdo',
    author_email = 'dingyeran@163.com',
    license='BSD',
    packages=find_packages(exclude=('tests', 'tests.*','Scrapyer.egg-info','extras','docs')),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': ['scrapy = scrapy.cmdline:execute']
    },
    # classifiers=[
    #     'Framework :: Scrapy',
    #     'Development Status :: 5 - Production/Stable',
    #     'Environment :: Console',
    #     'Intended Audience :: Developers',
    #     'License :: OSI Approved :: BSD License',
    #     'Operating System :: OS Independent',
    #     'Programming Language :: Python',
    #     'Programming Language :: Python :: 3',
    #     'Programming Language :: Python :: 3.6',
    #     'Programming Language :: Python :: 3.7',
    #     'Programming Language :: Python :: 3.8',
    #     'Programming Language :: Python :: 3.9',
    #     'Programming Language :: Python :: 3.10',
    #     'Programming Language :: Python :: Implementation :: CPython',
    #     'Programming Language :: Python :: Implementation :: PyPy',
    #     'Topic :: Internet :: WWW/HTTP',
    #     'Topic :: Software Development :: Libraries :: Application Frameworks',
    #     'Topic :: Software Development :: Libraries :: Python Modules',
    # ],
    python_requires='>=3.6',
    install_requires=install_requires,
    extras_require=extras_require,
)



from scrapy import signals
import base64,random
# from Team_Public.configs import dnot_use_proxy_spiders
from scrapy.utils.conf import get_config


class ProxyMiddleware(object):
    # Not all methods need to be defined. If a method is not defined,
    # scrapy acts as if the downloader middleware does not modify the
    # passed objects.
    def __init__(self, proxies):
        self.proxies = proxies

    @classmethod
    def from_crawler(cls, crawler): 
        proxies = [{i:j for i,j in get_config()[proxy].items()} for proxy in get_config().sections() if proxy.startswith("proxy_no")]
        s = cls(proxies=proxies)
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def process_request(self, request, spider):
        # Called for each request that goes through the downloader
        # middleware.
        proxy = random.choice(self.proxies)
        proxyUser = proxy.get("proxy_user")
        proxyPass = proxy.get("proxy_pass")
        proxyServer = proxy.get("proxy_server")
        if proxyUser and proxyPass and proxyServer:
            proxyAuth = "Basic " + base64.urlsafe_b64encode(bytes((proxyUser + ":" + proxyPass), "ascii")).decode("utf8")
            request.meta["proxy"] = proxyServer
            request.headers["Proxy-Authorization"] = proxyAuth

    def process_response(self, request, response, spider):
        return response

    def process_exception(self, request, exception, spider):
        pass

    def spider_opened(self, spider):
        spider.logger.info('Spider opened: %s' % spider.name)



from scrapy.utils.conf import get_config
import pymysql
from pymysql.err import IntegrityError,OperationalError

import logging
import json,os,re

class MySQLPipeline:
    def __init__(self, host, database, user, password, port):
        self.host = host
        self.database = database
        self.user = user
        self.password = password
        self.port = port

    @classmethod
    def from_crawler(cls, crawler):
        # 判断运行环境<根据环境变量中是否配置IF_PROD=True,或 测试正式环境settings["IF_PROD"] = True>
        section_select ="mysql_cfg_prod" if os.environ.get('IF_PROD') == "True" or crawler.settings.get('IF_PROD')\
             == True or get_config().get("settings","IF_PROD",fallback="False")=="True" else "mysql_cfg_dev"
        # 取 scrapy.cfg 参与配置
        return cls(host=get_config().get(section=section_select,option="MYSQL_HOST",fallback='') or crawler.settings.get('MYSQL_HOST'),
            database=get_config().get(section=section_select,option='MYSQL_DATABASE',fallback='') or crawler.settings.get('MYSQL_DATABASE'),
            user=get_config().get(section=section_select,option='MYSQL_USER',fallback='') or crawler.settings.get('MYSQL_USER'),
            password=get_config().get(section=section_select,option='MYSQL_PASSWORD',fallback='') or crawler.settings.get('MYSQL_PASSWORD'),
            port=get_config().getint(section=section_select,option='MYSQL_PORT',fallback='') or crawler.settings.get('MYSQL_PORT'),
        )

    def open_spider(self, spider):
        self.db = pymysql.connect(host=self.host, user=self.user, password=self.password, database=self.database, charset='utf8', port=self.port)
        self.cursor = self.db.cursor()

    def close_spider(self, spider):
        self.db.close()

    def process_item(self, item, spider):
        data = dict(item)
        if data.get("title", None):  # brand_id :: 出去型号两边空格
            data["title"] = data["title"].strip()
        if data.get("sources", None):  # sources :: 出去sources两边空格
            data["sources"] = data["sources"].strip()
        if data.get("brand_name", None):  # brand_name :: 出去品牌两边空格
            data["brand_name"] = data["brand_name"].strip()
        if data.get("category_name", None):  # category_name :: 去两端空格
            data["category_name"] = data["category_name"].strip()
        if data.get("category_id", None):  # category_id :: 出去category_id两边空格
            data["category_id"] = data["category_id"].strip()
        if data.get("p_category_id", None):  # p_category_id :: 去两端空格
            data["p_category_id"] = data["p_category_id"].strip()
        if data.get("p_category_name", None):  # p_category_name :: 去两端空格
            data["p_category_name"] = data["p_category_name"].strip()
        if data.get("brand_id", None):  # brand_id :: int
            data["brand_id"] = int(data["brand_id"])
        if data.get("level", None):  # level :: int
            data["level"] = int(data["level"])
        if data.get("list_json", None):  # list_json :: json
            if isinstance(data["list_json"], dict):
                data["list_json"] = json.dumps(dict(data["list_json"],ensure_ascii=False))
        if data.get('sources',None):  # source 来源转小写
            data['sources'] = data['sources'].strip().lower()
        if data.get("min_work_tp",None):  # min_work_tp :: int
            data["min_work_tp"] = int(data["min_work_tp"])
        if data.get("max_work_tp",None):  # max_work_tp :: int
            data["max_work_tp"] = int(data["max_work_tp"])
        if data.get("moq", None):  # moq :: int
            data["moq"] = int(data["moq"].replace(",",''))
        if data.get("mpq", None):  # mpq :: int
            data["mpq"] = int(data["mpq"].replace(",",''))
        if data.get("rough_weight", None):  # rough_weight :: float #注意单位是g
            data["rough_weight"] = int(data["rough_weight"])
        keys = ', '.join(data.keys())
        values = ', '.join(['% s'] * len(data))
        sql = 'insert into % s (% s) values (% s)' % (item.table, keys, values)
        msg = data.get("title", "") or data.get("category_name","") or data.get("url","") or data.get("sources","") # or data.get("")
        try:
            self.db.ping(reconnect=True)
            self.cursor.execute(sql, tuple(data.values()))
            self.db.commit()
            logging.debug(f'Crawl {msg} done.' )
        except IntegrityError:
            self.db.rollback()
            logging.info(f'去重 {msg} Skip .') 
        except OperationalError as e:
            logging.warning(e.args[1])
            column = re.findall("Unknown column '(.*?)' in 'field list'",e.args[1])[0]
            logging.warning(f"字段 {column} 在数据库中未创建，请自查后重试！ ")
        return item


from scrapy.xcc_downloadermiddlewares.fake_ua import choice_brower_model

class RandomUserAgent(object):
    DEFAULT_BROWSER = 'chrome'

    def __init__(self, uabrowser):
        self.browser = uabrowser  # if self.uabrowser else self.DEFAULT_BROWSER

    @classmethod
    def from_crawler(cls, crawler):
        ub = crawler.settings.get('UABROWSER')
        return cls(uabrowser=ub)

    def process_request(self, request, spider):
        request.headers["User-Agent"] = choice_brower_model(browser=self.browser)



import shutil
import tempfile
import unittest

import queuelib

from scrapy.squeues import (
    PickleFifoDiskQueue,
    PickleLifoDiskQueue,
    MarshalFifoDiskQueue,
    MarshalLifoDiskQueue,
    FifoMemoryQueue,
    LifoMemoryQueue,
)
from scrapy.http import Request
from scrapy.spiders import Spider
from scrapy.utils.test import get_crawler


"""
Queues that handle requests
"""


class BaseQueueTestCase(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix="scrapy-queue-tests-")
        self.qpath = self.tempfilename()
        self.qdir = self.mkdtemp()
        self.crawler = get_crawler(Spider)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def tempfilename(self):
        with tempfile.NamedTemporaryFile(dir=self.tmpdir) as nf:
            return nf.name

    def mkdtemp(self):
        return tempfile.mkdtemp(dir=self.tmpdir)


class RequestQueueTestMixin:
    def queue(self):
        raise NotImplementedError()

    def test_one_element_with_peek(self):
        if not hasattr(queuelib.queue.FifoMemoryQueue, "peek"):
            raise unittest.SkipTest("The queuelib queues do not define peek")
        q = self.queue()
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.peek())
        self.assertIsNone(q.pop())
        req = Request("http://www.example.com")
        q.push(req)
        self.assertEqual(len(q), 1)
        self.assertEqual(q.peek().url, req.url)
        self.assertEqual(q.pop().url, req.url)
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.peek())
        self.assertIsNone(q.pop())
        q.close()

    def test_one_element_without_peek(self):
        if hasattr(queuelib.queue.FifoMemoryQueue, "peek"):
            raise unittest.SkipTest("The queuelib queues define peek")
        q = self.queue()
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.pop())
        req = Request("http://www.example.com")
        q.push(req)
        self.assertEqual(len(q), 1)
        with self.assertRaises(NotImplementedError, msg="The underlying queue class does not implement 'peek'"):
            q.peek()
        self.assertEqual(q.pop().url, req.url)
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.pop())
        q.close()


class FifoQueueMixin(RequestQueueTestMixin):
    def test_fifo_with_peek(self):
        if not hasattr(queuelib.queue.FifoMemoryQueue, "peek"):
            raise unittest.SkipTest("The queuelib queues do not define peek")
        q = self.queue()
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.peek())
        self.assertIsNone(q.pop())
        req1 = Request("http://www.example.com/1")
        req2 = Request("http://www.example.com/2")
        req3 = Request("http://www.example.com/3")
        q.push(req1)
        q.push(req2)
        q.push(req3)
        self.assertEqual(len(q), 3)
        self.assertEqual(q.peek().url, req1.url)
        self.assertEqual(q.pop().url, req1.url)
        self.assertEqual(len(q), 2)
        self.assertEqual(q.peek().url, req2.url)
        self.assertEqual(q.pop().url, req2.url)
        self.assertEqual(len(q), 1)
        self.assertEqual(q.peek().url, req3.url)
        self.assertEqual(q.pop().url, req3.url)
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.peek())
        self.assertIsNone(q.pop())
        q.close()

    def test_fifo_without_peek(self):
        if hasattr(queuelib.queue.FifoMemoryQueue, "peek"):
            raise unittest.SkipTest("The queuelib queues do not define peek")
        q = self.queue()
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.pop())
        req1 = Request("http://www.example.com/1")
        req2 = Request("http://www.example.com/2")
        req3 = Request("http://www.example.com/3")
        q.push(req1)
        q.push(req2)
        q.push(req3)
        with self.assertRaises(NotImplementedError, msg="The underlying queue class does not implement 'peek'"):
            q.peek()
        self.assertEqual(len(q), 3)
        self.assertEqual(q.pop().url, req1.url)
        self.assertEqual(len(q), 2)
        self.assertEqual(q.pop().url, req2.url)
        self.assertEqual(len(q), 1)
        self.assertEqual(q.pop().url, req3.url)
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.pop())
        q.close()


class LifoQueueMixin(RequestQueueTestMixin):
    def test_lifo_with_peek(self):
        if not hasattr(queuelib.queue.FifoMemoryQueue, "peek"):
            raise unittest.SkipTest("The queuelib queues do not define peek")
        q = self.queue()
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.peek())
        self.assertIsNone(q.pop())
        req1 = Request("http://www.example.com/1")
        req2 = Request("http://www.example.com/2")
        req3 = Request("http://www.example.com/3")
        q.push(req1)
        q.push(req2)
        q.push(req3)
        self.assertEqual(len(q), 3)
        self.assertEqual(q.peek().url, req3.url)
        self.assertEqual(q.pop().url, req3.url)
        self.assertEqual(len(q), 2)
        self.assertEqual(q.peek().url, req2.url)
        self.assertEqual(q.pop().url, req2.url)
        self.assertEqual(len(q), 1)
        self.assertEqual(q.peek().url, req1.url)
        self.assertEqual(q.pop().url, req1.url)
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.peek())
        self.assertIsNone(q.pop())
        q.close()

    def test_lifo_without_peek(self):
        if hasattr(queuelib.queue.FifoMemoryQueue, "peek"):
            raise unittest.SkipTest("The queuelib queues do not define peek")
        q = self.queue()
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.pop())
        req1 = Request("http://www.example.com/1")
        req2 = Request("http://www.example.com/2")
        req3 = Request("http://www.example.com/3")
        q.push(req1)
        q.push(req2)
        q.push(req3)
        with self.assertRaises(NotImplementedError, msg="The underlying queue class does not implement 'peek'"):
            q.peek()
        self.assertEqual(len(q), 3)
        self.assertEqual(q.pop().url, req3.url)
        self.assertEqual(len(q), 2)
        self.assertEqual(q.pop().url, req2.url)
        self.assertEqual(len(q), 1)
        self.assertEqual(q.pop().url, req1.url)
        self.assertEqual(len(q), 0)
        self.assertIsNone(q.pop())
        q.close()


class PickleFifoDiskQueueRequestTest(FifoQueueMixin, BaseQueueTestCase):
    def queue(self):
        return PickleFifoDiskQueue.from_crawler(crawler=self.crawler, key="pickle/fifo")


class PickleLifoDiskQueueRequestTest(LifoQueueMixin, BaseQueueTestCase):
    def queue(self):
        return PickleLifoDiskQueue.from_crawler(crawler=self.crawler, key="pickle/lifo")


class MarshalFifoDiskQueueRequestTest(FifoQueueMixin, BaseQueueTestCase):
    def queue(self):
        return MarshalFifoDiskQueue.from_crawler(crawler=self.crawler, key="marshal/fifo")


class MarshalLifoDiskQueueRequestTest(LifoQueueMixin, BaseQueueTestCase):
    def queue(self):
        return MarshalLifoDiskQueue.from_crawler(crawler=self.crawler, key="marshal/lifo")


class FifoMemoryQueueRequestTest(FifoQueueMixin, BaseQueueTestCase):
    def queue(self):
        return FifoMemoryQueue.from_crawler(crawler=self.crawler)


class LifoMemoryQueueRequestTest(LifoQueueMixin, BaseQueueTestCase):
    def queue(self):
        return LifoMemoryQueue.from_crawler(crawler=self.crawler)
